# (c) Kevin Dunn, 2019-2021. MIT License.

import pathlib


import pytest
from pytest import approx
import pandas as pd
import numpy as np

from sklearn.cross_decomposition import PLSRegression
from process_improve.multivariate.methods import (
    epsqrt,
    center,
    scale,
    quick_regress,
    ssq,
    PCA,
    MCUVScaler,
    PLS,
    SpecificationWarning,
)


def test_PCA_SPE_limits():
    """
    Simulate data and see if SPE limit cuts off at 5%.
    """
    N = 1000
    repeats = 50
    outliers_95 = []
    outliers_99 = []
    for k in range(repeats):

        # The desired mean values of the sample.
        mu = np.array([0.0, 0.0, 0.0])

        # The desired covariance matrix.
        r = np.array([[5.20, -4.98, -1.00], [-4.98, 5.50, 2.94], [-1.00, 2.94, 2.77]])

        X = pd.DataFrame(np.random.multivariate_normal(mu, r, size=N))
        scaler = MCUVScaler().fit(X)
        mcuv = scaler.fit_transform(X)

        A = 2
        pca = PCA(n_components=A).fit(mcuv)
        SPE_limit_95 = pca.SPE_limit(0.95)
        SPE_limit_99 = pca.SPE_limit(0.99)

        outliers_95.append(
            (pca.squared_prediction_error.iloc[:, A - 1] > SPE_limit_95).sum()
        )
        outliers_99.append(
            (pca.squared_prediction_error.iloc[:, A - 1] > SPE_limit_99).sum()
        )

    assert np.mean(outliers_95) == approx(0.05 * N, rel=0.1)
    assert np.mean(outliers_99) == approx(0.01 * N, rel=0.1)


def test_PCA_foods():
    """
    Arrays with no variance should not be able to have variance extracted.
    """

    foods = pd.read_csv("https://openmv.net/file/food-texture.csv").drop(
        [
            "Unnamed: 0",
        ],
        axis=1,
    )
    scaler = MCUVScaler().fit(foods)
    foods_mcuv = scaler.fit_transform(foods)

    A = 2
    pca = PCA(n_components=A).fit(foods_mcuv)

    assert np.linalg.norm(
        np.diag(pca.x_scores.T @ pca.x_scores) / (pca.N - 1) - pca.explained_variance_
    ) == approx(0, abs=epsqrt)

    T2_limit_95 = pca.T2_limit(0.95)
    assert T2_limit_95 == approx(6.64469, rel=1e-3)

    pca.SPE_limit(0.95)

    ellipse_x, ellipse_y = pca.ellipse_coordinates(1, 2, 0.95, 100)
    assert ellipse_x[-1] == approx(4.48792, rel=1e-5)
    assert ellipse_y[-1] == approx(0, rel=1e-7)


@pytest.fixture
def fixture_kamyr_data_missing_value():
    folder = (
        pathlib.Path(__file__).parents[1]
        / "process_improve"
        / "datasets"
        / "multivariate"
    )
    return pd.read_csv(
        folder / "kamyr.csv",
        index_col=None,
        header=None,
    )


def test_PCA_missing_data(fixture_kamyr_data_missing_value):

    X_mcuv = MCUVScaler().fit_transform(fixture_kamyr_data_missing_value)

    # Build the model
    A = 2
    pca = PCA(n_components=A)
    assert pca.missing_data_settings is None

    # Check that default missing data options were used
    model = pca.fit(X_mcuv)
    assert isinstance(model.missing_data_settings, dict)
    assert "md_tol" in model.missing_data_settings

    assert np.linalg.norm(
        (model.loadings.T @ model.loadings) - np.eye(model.A)
    ) == approx(0, abs=1e-2)


@pytest.fixture
def fixture_mv_utilities():
    """
    Multivariate methods depend on an internal regression and Sum of Squares
    calculations. This code tests those crucial steps.
    """
    x = np.asarray([1, 2, 3, 4, 5, 6]).reshape(6, 1)
    Y = np.asarray(
        [
            [1, 2, 3, 4, 5, 6],
            [6, 5, 4, 3, 2, 1],
            [0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1],
            [1, np.NaN, 3, np.NaN, 5, np.NaN],
        ]
    )
    Y = Y.T
    return x, Y


def test_ssq(fixture_mv_utilities):
    x, _ = fixture_mv_utilities
    assert (1 + 2 * 2 + 3 * 3 + 4 * 4 + 5 * 5 + 6 * 6) == approx(ssq(x), abs=1e-9)


def test_quick_regress(fixture_mv_utilities):
    x, Y = fixture_mv_utilities
    out = quick_regress(Y, x).ravel()
    assert 1 == approx(out[0], abs=1e-9)
    assert 0.61538462 == approx(out[1], abs=1e-8)
    assert 0 == approx(out[2], abs=1e-9)

    # Checked against R: summary(lm(c(1,1,1,1,1,1) ~ seq(6) + 0))
    assert 0.23077 == approx(out[3], abs=1e-6)

    # Checked against what is expected: (1 + 3^2 + 5^2)/(1 + 3^2 + 5^2)
    assert 1.0 == approx(out[4], abs=1e-14)


@pytest.fixture
def fixture_tablet_spectra_data():
    """
    Verifies the PCA model for the case of no missing data.
    # R code:
    # -------
    # Read large data file
    file <- 'http://openmv.net/file/tablet-spectra.csv'
    spectra <- read.csv(file, header = FALSE, row.names = 1)
    # Only extract 4 components, but
    # center and scale the data before
    # calculation the components
    model.pca <- prcomp(spectra,
                        center = TRUE,
                        scale =TRUE,
                        rank. = 4)
    summary(model.pca)

    Importance of first k=4 (out of 460) components:
                               PC1     PC2     PC3     PC4
    Standard deviation     21.8835 10.9748 3.60075 3.27081
    Proportion of Variance  0.7368  0.1853 0.01995 0.01646
    Cumulative Proportion   0.7368  0.9221 0.94200 0.95846


    # T' * T on the scores matrix T:
    t(model.pca$x) %*% model.pca$x

                  PC1          PC2           PC3          PC4
    PC1  2.198092e+05 6.885159e-11 -1.134026e-11 3.454659e-11
    PC2  6.885159e-11 5.528459e+04  2.042206e-10 5.821477e-11
    PC3 -1.134026e-11 2.042206e-10  5.951125e+03 7.815970e-13
    PC4  3.454659e-11 5.821477e-11  7.815970e-13 4.910481e+03
    """
    folder = (
        pathlib.Path(__file__).parents[1]
        / "process_improve"
        / "datasets"
        / "multivariate"
    )
    spectra = pd.read_csv(
        folder / "tablet-spectra.csv",
        index_col=0,
        header=None,
    )

    # Ignoring values < 1E-8 (round them to zero) from the R output above.
    known_scores_covar = np.array(
        [
            [2.198092e05, 0, 0, 0],
            [0, 5.528459e04, 0, 0],
            [0, 0, 5.951125e03, 0],
            [0, 0, 0, 4.910481e03],
        ]
    )
    return spectra, known_scores_covar


def test_MCUV_centering(fixture_tablet_spectra_data):
    """
    Mean centering of the testing data.
    """

    spectra, _ = fixture_tablet_spectra_data
    X_mcuv = MCUVScaler().fit_transform(spectra)
    assert 0.0 == approx(np.max(np.abs(X_mcuv.mean(axis=0))), rel=1e-9)


def test_MCUV_scaling(fixture_tablet_spectra_data):
    """Scaling by standard deviation."""

    spectra, _ = fixture_tablet_spectra_data
    X_mcuv = MCUVScaler().fit_transform(spectra)

    assert 1 == approx(np.min(np.abs(X_mcuv.std(axis=0))), 1e-10)
    assert 1 == approx(X_mcuv.std(), 1e-10)


def test_PCA_tablet_spectra(fixture_tablet_spectra_data):
    r"""
    PCA characteristics:

    1. model's loadings must be orthogonal if there are no missing data.
        P.T * P = I
    2. model's loadings must be of unit length (norm = 1.0)
        P.T * P = I
    3. model's scores must be orthogonal
        T.T * T is a diagonal matrix when there's no missing data
    4. each earlier score's variance must be >= variance of later score


    PCA models have the following properties:

    * :math:`p_i'p_j' = 0` for :math:`i\neq j`; i.e. :math:`p_i \perp p_j`
    * :math:`t_i't_j' = 0` for :math:`i\neq j`; i.e. :math:`t_i \perp t_j`
    * :math:`P'P = I_A` when extracting :math:`A` components
    * :math:`P_{all} \text{ is a } \min(N,K) \times \min(N,K)` matrix, for all components
    * :math:`T_{all} \text{ is a } \min(N,K) \times \min(N,K)` matrix, for all components
    (it is just a rearrangement of X)
    * :math:`\text{SVD}(X): UDV' = X` and :math:`V' = P'` and :math:`UD = T`
    """

    spectra, known_scores_covar = fixture_tablet_spectra_data

    # Number of components to calculate
    model = PCA(n_components=2)
    model.fit(scale(center(spectra)))

    # P'P = identity matrix of size A x A
    orthogonal_check = model.loadings.T @ model.loadings
    assert 0.0 == approx(np.linalg.norm(orthogonal_check - np.eye(model.A)), rel=1e-9)

    # Check the R2 value against the R software output
    assert model.R2cum[0] == approx(0.7368, rel=1e-3)
    assert model.R2cum[1] == approx(0.9221, rel=1e-2)

    # Unit length: actually checked above, via subtraction with I matrix.
    # Check if scores are orthogonal
    scores_covar = model.x_scores.T @ model.x_scores
    for i in range(model.A):
        for j in range(model.A):

            # Technically not need, but more explict this way.
            if i == j:
                assert scores_covar.iloc[i, j] == approx(
                    known_scores_covar[i, j], rel=1e-2
                )
            else:
                assert scores_covar.iloc[i, j] == approx(
                    known_scores_covar[i, j], abs=1e-4
                )

                if i >= 1:
                    assert scores_covar.iloc[j, j] > scores_covar.iloc[i, i]

    # Check the model against an SVD: this raw data set has no missing
    # data, so the SVD should be faster and more accurate than NIPALS
    autoscaled_X = scale(center(spectra))
    u, s, v = np.linalg.svd(autoscaled_X)

    loadings_delta = np.linalg.norm(
        np.abs(v[0 : model.A, :]) - np.abs(model.loadings.T)
    )
    assert loadings_delta == approx(0, abs=1e-8)

    # It is not possible, it seems, to get the scores to match the SVD
    # scores. Numerical error?


def test_PCA_errors_no_variance_to_start():
    """
    Arrays with no variance should seem to work, but should have no variability explained.
    """
    K, N, A = 17, 12, 5
    data = pd.DataFrame(np.zeros((N, K)))
    model = PCA(n_components=A)
    # with pytest.raises(RuntimeError):
    model.fit(data)
    assert np.sum(model.x_scores.values, axis=None) == approx(0, abs=epsqrt)
    assert model.R2cum.sum() == approx(0, abs=epsqrt)
    assert np.isnan(model.R2cum[A - 1])


def test_PCA_invalid_calls():
    """
    Tests various invalid calls, and corresponding error messages.
    """
    K, N, A = 4, 3, 5
    data = pd.DataFrame(np.random.uniform(low=-1, high=1, size=(N, K)))
    with pytest.warns(
        SpecificationWarning,
        match=r"The requested number of components is more than can be computed from data(.*)",
    ):
        model = PCA(n_components=A)
        model.fit(data)

    data.iloc[0, 0] = np.nan
    with pytest.raises(AssertionError, match="Tolerance must exceed machine precision"):
        _ = PCA(
            n_components=A, missing_data_settings=dict(md_method="nipals", md_tol=0)
        ).fit(data)

    with pytest.raises(
        AssertionError, match=r"Missing data method is not recognized(.*)"
    ):
        _ = PCA(n_components=A, missing_data_settings={"md_method": "SCP"}).fit(data)

    # TODO: replace with a check to ensure the data is in a DataFrame.
    # from scipy.sparse import csr_matrix
    # sparse_data = csr_matrix([[1, 2, 0], [0, 0, 3], [4, 0, 5]])
    # with pytest.raises(TypeError, match="This PCA class does not support sparse input."):
    #     model = PCA(n_components=2)
    #     model.fit(sparse_data)


def test_PCA_no_more_variance():
    """
    Create a rank 2 matrix and it should fail on the 3rd component.
    """

    K = 17
    N = 12
    A = 3
    T = np.random.uniform(low=-1, high=1, size=(N, 2))
    P = np.random.uniform(low=-1, high=1, size=(K, 2))
    X = T @ P.T
    meanX = X.mean(axis=0)
    stdX = X.std(axis=0, ddof=0)
    X = pd.DataFrame((X - meanX) / stdX)
    _ = PCA(n_components=A)

    # with pytest.raises(RuntimeError):
    #    m.fit(X)
    # TODO: check that the m.R2[2] (3rd PC is zero.)


def test_PCA_columns_with_no_variance():
    """
    Create a column with no variance. That column's loadings should be 0.
    """
    K = 14
    N = 29
    A = 4
    cols_with_no_variance = [10, 3]
    T = np.random.uniform(low=-1, high=1, size=(N, A))
    P = np.random.uniform(low=-1, high=1, size=(K, A))
    X = T @ P.T
    meanX = X.mean(axis=0)
    stdX = X.std(axis=0, ddof=0)
    X = pd.DataFrame((X - meanX) / stdX)
    X.iloc[:, cols_with_no_variance] = 0

    m = PCA(n_components=2)
    m.fit(X)

    # `loadings` is a K by A matrix.  Check sum of loadings in rows with
    # no variance must be zero
    assert np.sum(np.abs(m.loadings.iloc[cols_with_no_variance, :].values)) == approx(
        0, abs=1e-14
    )
    # The loadings must still be orthonormal though:
    assert np.sum(np.identity(m.A) - m.loadings.values.T @ m.loadings.values) == approx(
        0, abs=1e-14
    )

    # Are scores orthogonal?
    covmatrix = m.x_scores.T @ m.x_scores
    covmatrix - np.diag(np.diag(covmatrix))
    (np.sum(np.abs(covmatrix - np.diag(np.diag(covmatrix))))).values == approx(
        0, abs=1e-6
    )


@pytest.fixture
def fixture_pca_PCA_Wold_etal_paper():
    """
    From the PCA paper by Wold, Esbensen and Geladi, 1987
    Principal Component Analysis, Chemometrics and Intelligent Laboratory
    Systems, v 2, p37-52; http://dx.doi.org/10.1016/0169-7439(87)80084-9
    """
    return pd.DataFrame(np.array([[3, 4, 2, 2], [4, 3, 4, 3], [5.0, 5, 6, 4]]))


def test_PCA_Wold_centering(fixture_pca_PCA_Wold_etal_paper):
    """
    Checks the centering step
    """
    out, centering = center(fixture_pca_PCA_Wold_etal_paper, extra_output=True)
    assert centering == approx([4, 4, 4, 3], rel=1e-8)


def test_PCA_Wold_scaling(fixture_pca_PCA_Wold_etal_paper):
    """
    Checks the scaling step. Page 40 of the above paper.
    """

    out, scaling = scale(
        center(fixture_pca_PCA_Wold_etal_paper), extra_output=True, ddof=1
    )
    assert scaling == approx([1, 1, 0.5, 1])


def test_PCA_Wold_model_results(fixture_pca_PCA_Wold_etal_paper):
    """
    Checks if the PCA model matches the results in the paper.
    """

    X_preproc = scale(center(fixture_pca_PCA_Wold_etal_paper))
    pca_1 = PCA(n_components=1)
    pca_1.fit(X_preproc.copy())

    # TODO: complete these tests

    # The remaining sum of squares, on page 43
    # SS_X = np.sum(pca_1["residuals"].values ** 2, axis=0)
    # self.assertTrue(
    #     np.all(compare_entries(SS_X, np.array([0.0551, 1.189, 0.0551, 0.0551]), 3))
    # )

    # # The residuals after 1 component
    # self.assertTrue(
    #     np.all(compare_entries(SS_X, np.array([0.0551, 1.189, 0.0551, 0.0551]), 3))
    # )

    # # With 2 components, the loadings are, page 40
    # P.T = [ 0.5410, 0.3493,  0.5410,  0.5410],
    #      [-0.2017, 0.9370, -0.2017, -0.2017]
    X_preproc = scale(center(fixture_pca_PCA_Wold_etal_paper))
    pca_2 = PCA(n_components=2)
    pca_2.fit(X_preproc)
    assert np.abs(pca_2.loadings.values[:, 0]) == approx(
        [0.5410, 0.3493, 0.5410, 0.5410], abs=1e-4
    )
    assert np.abs(pca_2.loadings.values[:, 1]) == approx(
        [0.2017, 0.9370, 0.2017, 0.2017], abs=1e-4
    )

    # Scores. The scaling is off here by a constant factor of 0.8165
    # assert np.all(pca_2.x_scores["1"] == approx([-1.6229, -0.3493, 1.9723], rel=1e-3))
    # assert np.all(pca_2.x_scores["2"] == approx([0.6051, -0.9370, 0.3319], rel=1e-4))

    # R2 values, given on page 43
    assert pca_2.R2.values == approx([0.831, 0.169], rel=1e-2)

    # SS values, on page 43
    # SS_X = np.sum(X_preproc ** 2, axis=0)
    # assert SS_X == approx([0.0, 0.0, 0.0, 0.0], abs=1e-9)

    # Testing data:
    # X_test = Block(np.array([[3, 4, 3, 4], [1, 2, 3, 4.0]]))
    # X.preprocess(X_test)
    # compare_entries(X_test.data, np.array([[-1, 0, -0.5, 1],
    # [-3, -2, -0.5, 1]])
    # #testing = PCA_model.apply(X_test)
    # compare_entries(testing.T, np.array([[-0.2075, 0.1009],
    # [-2.0511, -1.3698]])


def test_PLS_properties_TODO():
    """
    TODO:
    diag(T.T * T) related to S
    W.T * W = I for PLS only
    P.T * W: ones on diagonal, zeros below diagonal
    W.T * R: ones on diagonal, zeros below diagonal
    R.T * P = ID
    """
    pass


@pytest.mark.skip(reason="API still has to be improved to handle this case")
def test_PLS_invalid_calls():
    """
    Tests various invalid calls, and corresponding error messages.
    """
    K, N, M, A = 4, 3, 2, 5
    dataX = pd.DataFrame(np.random.uniform(low=-1, high=1, size=(N, K)))
    dataY = pd.DataFrame(np.random.uniform(low=-1, high=1, size=(N, M)))
    with pytest.raises(
        ValueError, match="Tolerance `tol`` must be between 1E-16 and 1.0"
    ):
        _ = PLS(n_components=A, tol=0)

    with pytest.raises(ValueError, match="Method 'SVDS' is not known."):
        _ = PLS(n_components=A, method="SVDS")

    with pytest.raises(ValueError, match="Missing data method 'SCP' is not known."):
        _ = PLS(n_components=A, md_method="SCP")

    with pytest.warns(
        SpecificationWarning, match=r"The requested number of components is (.*)"
    ):
        model = PLS(
            n_components=A,
        )
        model.fit(dataX, dataY)

    from scipy.sparse import csr_matrix

    sparse_data = csr_matrix([[1, 2], [0, 3], [4, 5]])
    with pytest.raises(
        TypeError, match="This PLS class does not support sparse input."
    ):
        model = PLS(n_components=2)
        model.fit(dataX, sparse_data)


@pytest.fixture
def fixture_PLS_model_SIMCA_1_component():
    """
    Simple model tested against Simca-P, version 14.1.
    Testing on 28 June 2020.

    When X and y are mean centered and scaled, the model should provide the loadings as listed here

    TODO: test against R

    X = matrix(c(41.1187, 21.2833, 21.1523,  0.2446, -0.0044, -0.131,  1.12,
                 41.7755, 22.0978, 21.1653,  0.3598,  0.1622, -0.9325, 1.01,
                 41.2568, 21.4873, 20.7407,  0.2536,  0.1635, -0.7467, 0.97,
                 41.5469, 22.2043, 20.4518,  0.6317,  0.1997, -1.7525, 0.83,
                 40.0234, 23.7399, 21.978,  -0.0534, -0.0158, -1.7619, 0.93,
                 39.9203, 21.9997, 21.5859, -0.1811,  0.089,  -0.4138, 1.02,
                 42.1886, 21.4891, 20.4427,  0.686,   0.1124, -1.0464, 0.91,
                 42.1454, 20.3803, 18.2327,  0.6607,  0.1291, -2.1476, 0.70,
                 42.272,  18.9725, 18.3763,  0.561,   0.0453, -0.5962, 1.26,
                 41.49,   18.603,  17.9978,  0.4872,  0.1198, -0.6052, 1.05,
                 41.5306, 19.1558, 18.2172,  0.6233,  0.1789, -0.9386, 0.95), nrow=11, byrow=T)

    data = data.frame(X)
    colnames(data) <- c("A","B","C", "D", "E", "F", "y")

    library(pls)
    model = plsr(y~., 1, data=data, method="simpls")
    plot(model)

    self.expected_y_predicted = [
        1.17475,
        0.930441,
        0.979066,
        0.773083,
        0.945719,
        1.10583,
        0.929441,
        0.796271,
        1.09379,
        1.0635,
        0.958111,
        ]
    """
    data = {}
    data["X"] = pd.DataFrame(
        np.array(
            [
                [
                    41.1187,
                    21.2833,
                    21.1523,
                    0.2446,
                    -0.0044,
                    -0.131,
                ],
                [
                    41.7755,
                    22.0978,
                    21.1653,
                    0.3598,
                    0.1622,
                    -0.9325,
                ],
                [
                    41.2568,
                    21.4873,
                    20.7407,
                    0.2536,
                    0.1635,
                    -0.7467,
                ],
                [
                    41.5469,
                    22.2043,
                    20.4518,
                    0.6317,
                    0.1997,
                    -1.7525,
                ],
                [
                    40.0234,
                    23.7399,
                    21.978,
                    -0.0534,
                    -0.0158,
                    -1.7619,
                ],
                [
                    39.9203,
                    21.9997,
                    21.5859,
                    -0.1811,
                    0.089,
                    -0.4138,
                ],
                [
                    42.1886,
                    21.4891,
                    20.4427,
                    0.686,
                    0.1124,
                    -1.0464,
                ],
                [
                    42.1454,
                    20.3803,
                    18.2327,
                    0.6607,
                    0.1291,
                    -2.1476,
                ],
                [
                    42.272,
                    18.9725,
                    18.3763,
                    0.561,
                    0.0453,
                    -0.5962,
                ],
                [
                    41.49,
                    18.603,
                    17.9978,
                    0.4872,
                    0.1198,
                    -0.6052,
                ],
                [
                    41.5306,
                    19.1558,
                    18.2172,
                    0.6233,
                    0.1789,
                    -0.9386,
                ],
            ]
        )
    )

    data["y"] = pd.DataFrame(
        np.array([1.12, 1.01, 0.97, 0.83, 0.93, 1.02, 0.91, 0.7, 1.26, 1.05, 0.95])
    )
    data["expected_y_predicted"] = [
        1.17475,
        0.930441,
        0.979066,
        0.773083,
        0.945719,
        1.10583,
        0.929441,
        0.796271,
        1.09379,
        1.0635,
        0.958111,
    ]
    data["loadings_P1"] = np.array(
        [
            -0.2650725,
            -0.2165038,
            0.08547913,
            -0.3954746,
            -0.4935882,
            0.7541404,
        ]
    )
    data["loadings_r1"] = np.array(
        [
            -0.04766187,
            -0.3137862,
            0.004006641,
            -0.238001,
            -0.4430451,
            0.8039384,
        ]
    )
    data["loadings_y_c1"] = 0.713365
    data["SDt"] = 1.19833
    data["R2X"] = 0.261641
    data["R2Y"] = 0.730769
    data["t1"] = np.array(
        [
            1.889566,
            -0.4481195,
            0.0171578,
            -1.953837,
            -0.3019302,
            1.230112,
            -0.4576912,
            -1.731961,
            1.114923,
            0.8251334,
            -0.1833536,
        ]
    )
    data["Tsq"] = np.array(
        [
            2.48638,
            0.1398399,
            0.0002050064,
            2.658398,
            0.0634829,
            1.053738,
            0.1458776,
            2.08891,
            0.8656327,
            0.4741239,
            0.02341113,
        ]
    )
    data["DModX"] = np.array(
        [
            0.367926,
            1.01727,
            0.970395,
            0.635592,
            2.36596,
            0.449567,
            0.645429,
            1.12458,
            0.520623,
            0.384443,
            0.764301,
        ]
    )
    data["Xavg"] = np.array(
        [41.38802, 21.03755, 20.03097, 0.3884909, 0.1072455, -1.006582]
    )
    data["Xws"] = 1 / np.array(
        [
            1.259059,
            0.628138,
            0.6594034,
            3.379028,
            13.8272,
            1.589986,
        ]
    )
    data["Yavg"] = 0.9772727
    data["Yws"] = 1 / 6.826007  # Simca-P uses inverse standard deviation
    data["A"] = 1
    data["conf"] = 0.95
    return data


def test_PLS_compare_sklearn_1_component(fixture_PLS_model_SIMCA_1_component):

    data = fixture_PLS_model_SIMCA_1_component

    plsmodel = PLSRegression(n_components=data["A"], scale="True")
    plsmodel.fit(data["X"], data["y"])

    # Check the pre-processing: sig figs have been taken as high as possible.
    assert plsmodel._x_mean == approx(data["Xavg"], abs=1e-5)
    assert plsmodel._x_std == approx(data["Xws"], abs=1e-6)
    assert plsmodel._y_mean == approx(data["Yavg"], abs=1e-7)
    assert plsmodel._y_std == approx(data["Yws"], abs=1e-8)

    # Extract the model parameters
    T = plsmodel.x_scores_
    P = plsmodel.x_loadings_
    assert T.ravel() == approx(data["t1"], abs=1e-5)
    assert np.std(T, ddof=1) == approx(data["SDt"], rel=1e-5)
    assert data["loadings_P1"].ravel() == approx(P.ravel(), rel=1e-5)
    assert data["loadings_r1"] == approx(plsmodel.x_weights_.ravel(), rel=1e-4)

    # Check the model's predictions
    t1_predict, y_pp = plsmodel.transform(data["X"], data["y"])
    assert data["t1"] == approx(t1_predict.ravel(), abs=1e-5)
    # assert y_pp == approx((data["y"] - data["Yavg"]) / data["Yws"], abs=1e-6)

    # Manually make the PLS prediction
    # X_check = data["X"].copy()
    # X_check_mcuv = (X_check - plsmodel._x_mean) / plsmodel._x_std
    # t1_predict_manually = X_check_mcuv @ plsmodel.x_weights_

    # TODO: fix the rest of this test. Not sure what the purpose of this test is anyway.

    # # Simca's C:
    # N = data["X"].shape[0]
    # simca_C = (y_pp.reshape(1, N) @ t1_predict) / (t1_predict.T @ t1_predict)
    # # assert simca_C == approx(data["loadings_y_c1"], 1e-6)
    # assert t1_predict_manually.values.ravel() == approx(t1_predict.ravel(), 1e-9)

    # # Deflate the X's:
    # X_check_mcuv -= t1_predict_manually @ plsmodel.x_loadings_.T
    # y_hat = t1_predict_manually @ simca_C
    # y_hat_rawunits = y_hat * plsmodel._y_std + plsmodel._y_mean
    # assert data["expected_y_predicted"] == approx(y_hat_rawunits.values.ravel(), abs=1e-5)

    # prediction_error = data["y"].values - y_hat_rawunits.values
    # R2_y = (data["y"].var(ddof=1) - prediction_error.var(ddof=1)) / data["y"].var(ddof=1)
    # assert R2_y == approx(data["R2Y"], abs=1e-6)


def test_PLS_compare_model_api(fixture_PLS_model_SIMCA_1_component):

    data = fixture_PLS_model_SIMCA_1_component
    plsmodel = PLS(n_components=data["A"])

    X_mcuv = MCUVScaler().fit(data["X"])
    Y_mcuv = MCUVScaler().fit(data["y"])

    # Check the pre-processing: sig figs have been taken as high as possible.
    assert X_mcuv.center_.values == approx(data["Xavg"], abs=1e-5)
    assert X_mcuv.scale_.values == approx(data["Xws"], abs=1e-6)
    assert Y_mcuv.center_.values == approx(data["Yavg"], abs=1e-7)
    assert Y_mcuv.scale_.values == approx(data["Yws"], abs=1e-8)

    # Extract the model parameters
    plsmodel.fit(X_mcuv.transform(data["X"]), Y_mcuv.transform(data["y"]))
    assert data["SDt"] == approx(np.std(plsmodel.x_scores, ddof=1), abs=1e-5)
    assert data["t1"] == approx(plsmodel.x_scores.values.ravel(), abs=1e-5)
    assert data["loadings_P1"] == approx(plsmodel.x_loadings.values.ravel(), abs=1e-5)
    assert data["loadings_r1"] == approx(plsmodel.x_weights.values.ravel(), abs=1e-6)

    assert data["expected_y_predicted"] == approx(
        Y_mcuv.inverse_transform(plsmodel.predictions).values.ravel(), abs=1e-5
    )
    assert data["R2Y"] == approx(plsmodel.R2cum, abs=1e-6)

    # Check the model's predictions
    state = plsmodel.predict(X_mcuv.transform(data["X"]))
    assert plsmodel.squared_prediction_error.values.ravel() == approx(
        state.squared_prediction_error.values, abs=1e-9
    )
    assert data["t1"] == approx(state.x_scores.values.ravel(), abs=1e-5)
    assert data["Tsq"] == approx(state.Hotellings_T2.values.ravel(), abs=1e-5)
    assert data["expected_y_predicted"] == approx(
        Y_mcuv.inverse_transform(state.y_hat).values.ravel(), abs=1e-5
    )


@pytest.fixture
def fixture_PLS_SIMCA_2_components():
    """
    Simple model tested against Simca-P, version 14.1.
    Testing on 02 July 2020.
    No missing data

    When X and y are mean centered and scaled, the model should provide the loadings listed here.
    """
    out = {}
    out["X"] = pd.DataFrame(
        np.array(
            [
                [
                    1.27472,
                    0.897732,
                    -0.193397,
                ],
                [
                    1.27472,
                    -1.04697,
                    0.264243,
                ],
                [
                    0.00166722,
                    1.26739,
                    1.06862,
                ],
                [
                    0.00166722,
                    -0.0826556,
                    -1.45344,
                ],
                [
                    0.00166722,
                    -1.46484,
                    1.91932,
                ],
                [
                    -1.27516,
                    0.849516,
                    -0.326239,
                ],
                [
                    -1.27516,
                    -1.06304,
                    0.317718,
                ],
                [
                    -0.000590006,
                    1.26739,
                    1.06862,
                ],
                [
                    -0.000590006,
                    -0.0826556,
                    -1.45344,
                ],
                [
                    -0.000590006,
                    -1.09519,
                    0.427109,
                ],
                [
                    -1.27516,
                    0.849516,
                    -0.326239,
                ],
                [
                    -1.27516,
                    -1.06304,
                    0.317718,
                ],
                [
                    1.27398,
                    0.897732,
                    -0.193397,
                ],
                [
                    1.27398,
                    -0.130872,
                    -1.4372,
                ],
            ]
        )
    )

    out["y"] = pd.DataFrame(
        np.array(
            [
                -0.0862851,
                -1.60162,
                0.823439,
                0.242033,
                -1.64304,
                1.59583,
                -0.301604,
                0.877623,
                0.274155,
                -0.967692,
                1.47491,
                -0.194163,
                0.097352,
                -0.590925,
            ]
        )
    )
    out["expected_y_predicted"] = [
        0.04587483,
        -1.671657,
        0.8337691,
        0.2326966,
        -1.622544,
        1.520105,
        -0.209406,
        0.8350853,
        0.2340128,
        -1.002176,
        1.520105,
        -0.209406,
        0.04630876,
        -0.552768,
    ]
    out["loadings_P"] = np.array(
        [[-0.3799977, -0.7815778], [0.8737038, -0.2803103], [-0.3314019, 0.55731]]
    )
    out["loadings_W"] = np.array(  # W
        [[-0.4839311, -0.7837874], [0.8361799, -0.2829775], [-0.2580969, 0.5528119]]
    )
    out["loadings_C"] = [1.019404, 0.1058565]
    out["SDt"] = [0.9724739, 1.098932]

    out["R2X"] = [
        0.3207782,
        0.4025633,
    ]  # cumulative: 32%, then 72% for second component
    out["R2Y"] = [0.9827625, 0.01353244]
    out["T"] = np.array(
        [
            [0.1837029, -1.335702],
            [-1.560534, -0.7636986],
            [0.7831483, 0.334647],
            [0.3052059, -0.7409231],
            [-1.721048, 1.246014],
            [1.411638, 0.7658994],
            [-0.3538088, 1.428992],
            [0.7842407, 0.3365611],
            [0.3062983, -0.7390091],
            [-1.025724, 0.4104719],
            [1.411638, 0.7658994],
            [-0.3538088, 1.428992],
            [0.184063, -1.335071],
            [-0.3550123, -1.803074],
        ]
    )
    # TODO: test against this still
    out["Tsq"] = np.array(
        [
            1.513014,
            3.05803,
            0.7412658,
            0.5530728,
            4.417653,
            2.592866,
            1.823269,
            0.74414,
            0.5514336,
            1.252029,
            2.592866,
            1.823269,
            1.511758,
            2.825334,
        ]
    )

    # TODO: test against this still
    out["DModX"] = np.array(
        [
            0.8796755,
            0.2482767,
            1.641307,
            1.350485,
            0.9410046,
            0.4101084,
            0.856736,
            1.640294,
            1.351499,
            0.2035393,
            0.4101084,
            0.856736,
            0.8793415,
            0.7906867,
        ]
    )
    out["A"] = 2
    return out


def test_PLS_sklearn_2_components(fixture_PLS_SIMCA_2_components):

    data = fixture_PLS_SIMCA_2_components

    plsmodel = PLSRegression(n_components=data["A"], scale=False)

    X_mcuv = MCUVScaler().fit_transform(data["X"])
    Y_mcuv = MCUVScaler().fit_transform(data["y"])

    plsmodel.fit(X_mcuv, Y_mcuv)

    # Extract the model parameters
    assert np.abs(data["T"]) == approx(np.abs(plsmodel.x_scores_), abs=1e-5)
    assert np.std(plsmodel.x_scores_, ddof=1, axis=0) == approx(data["SDt"], abs=1e-6)
    assert np.abs(data["loadings_P"]) == approx(np.abs(plsmodel.x_loadings_), abs=1e-5)
    assert np.abs(data["loadings_W"]) == approx(np.abs(plsmodel.x_weights_), abs=1e-5)


def test_PLS_compare_API(fixture_PLS_SIMCA_2_components):
    data = fixture_PLS_SIMCA_2_components

    plsmodel = PLS(n_components=data["A"])

    X_mcuv = MCUVScaler().fit(data["X"])
    Y_mcuv = MCUVScaler().fit(data["y"])
    plsmodel.fit(X_mcuv.transform(data["X"]), Y_mcuv.transform(data["y"]))

    # Extract the model parameters
    assert data["SDt"] == approx(np.std(plsmodel.x_scores, ddof=1, axis=0), abs=1e-6)
    assert np.abs(data["T"]) == approx(np.abs(plsmodel.x_scores), abs=1e-5)
    assert np.abs(data["loadings_P"]) == approx(np.abs(plsmodel.x_loadings), abs=1e-5)
    assert np.abs(data["loadings_W"]) == approx(np.abs(plsmodel.x_weights), abs=1e-5)
    assert Y_mcuv.inverse_transform(plsmodel.predictions).values == approx(
        data["expected_y_predicted"], abs=1e-5
    )
    assert sum(data["R2Y"]) == approx(plsmodel.R2cum.values[-1], abs=1e-7)

    # Check the model's predictions
    state = plsmodel.predict(X_mcuv.transform(data["X"]))
    # TODO: a check on SPE vs Simca-P. Here we are doing a check between the SPE from the
    # model building, to model-using, but not against an external library.
    assert plsmodel.squared_prediction_error.iloc[:, -1].values == approx(
        state.squared_prediction_error, abs=1e-10
    )
    assert data["Tsq"] == approx(state.Hotellings_T2, abs=1e-5)
    assert data["expected_y_predicted"] == approx(
        Y_mcuv.inverse_transform(state.y_hat).values.ravel(), abs=1e-5
    )
    assert np.abs(data["T"]) == approx(np.abs(state.x_scores), abs=1e-5)


@pytest.fixture
def fixture_PLS_LDPE_example():
    """
    No missing data.
    Source: https://openmv.net/info/ldpe

    Data from a low-density polyethylene production process.
    There are 14 process variables and 5 quality variables (last 5 columns).
    More details: http://dx.doi.org/10.1002/aic.690400509
    The first 50 observations are from common-cause (normal) operation, while the last 4 show a
    process fault developing: the impurity level in the ethylene feed in both zones is increasing.

    Tin: inlet temperature to zone 1 of the reactor [K]
    Tmax1: maximum temperature along zone 1 [K]
    Tout1: outlet temperature from zone 1 [K]
    Tmax2: maximum temperature along zone 2 [K]
    Tout2: outlet temperature from zone 2 [K]
    Tcin1: temperature of inlet coolant to zone [K]
    Tcin2: temperature of inlet coolant to zone 2 [K]
    z1: percentage along zone 1 where Tmax1 occurs [%]
    z2: percentage along zone 2 where Tmax2 occurs [%]
    Fi1: flow rate of initiators to zone 1 [g/s]
    Fi2: flow rate of initiators to zone 2 [g/s]np.abs(state.scores)
    Fs1: flow rate of solvent to zone 1 [% of ethylene]
    Fs2: flow rate of solvent to zone 2 [% of ethylene]
    Press: pressure in the reactor [atm]
    ------
    Conv: quality variable: cumulative conversion
    Mn: quality variable: number average molecular weight
    Mw: quality variable: weight average molecular weight
    LCB: quality variable: long chain branching per 1000 C atoms
    SCB: quality variable: short chain branching per 1000 C atoms

    N = 54
    K = 14
    M = 5
    A = 6
    """
    out = {}
    folder = (
        pathlib.Path(__file__).parents[1]
        / "process_improve"
        / "datasets"
        / "multivariate"
    )
    values = pd.read_csv(
        folder / "LDPE" / "LDPE.csv",
        index_col=0,
    )
    out["expected_T"] = pd.read_csv(folder / "LDPE" / "T.csv", header=None)
    out["expected_P"] = pd.read_csv(folder / "LDPE" / "P.csv", header=None)
    out["expected_W"] = pd.read_csv(folder / "LDPE" / "W.csv", header=None)
    out["expected_C"] = pd.read_csv(folder / "LDPE" / "C.csv", header=None)
    out["expected_U"] = pd.read_csv(folder / "LDPE" / "U.csv", header=None)
    out["expected_Hotellings_T2_A3"] = pd.read_csv(
        folder / "LDPE" / "Hotellings_T2_A3.csv",
        header=None,
    )
    out["expected_Hotellings_T2_A6"] = pd.read_csv(
        folder / "LDPE" / "Hotellings_T2_A6.csv",
        header=None,
    )
    out["expected_Yhat_A6"] = pd.read_csv(
        folder / "LDPE" / "Yhat_A6.csv",
        header=None,
    )
    out["expected_SD_t"] = np.array(
        [1.872539, 1.440642, 1.216218, 1.141096, 1.059435, 0.9459715]
    )
    out["expected_T2_lim_95_A6"] = 15.2017
    out["expected_T2_lim_99_A6"] = 21.2239
    out["X"] = values.iloc[:, :14]
    out["Y"] = values.iloc[:, 14:]
    assert out["X"].shape == approx([54, 14])
    assert out["Y"].shape == approx([54, 5])
    out["A"] = 6
    return out


def test_PLS_SIMCA_LDPE(fixture_PLS_LDPE_example):
    """Unit test for LDPE case study.

    Parameters
    ----------
    PLS_model_SIMCA_LDPE_example : dict
        Dictionary of raw data and expected outputs from the PLS model.
    """
    data = fixture_PLS_LDPE_example
    plsmodel = PLS(n_components=data["A"])

    X_mcuv = MCUVScaler().fit(data["X"])
    Y_mcuv = MCUVScaler().fit(data["Y"])
    plsmodel.fit(X_mcuv.transform(data["X"]), Y_mcuv.transform(data["Y"]))

    # Can only get these to very loosely match
    assert data["expected_T2_lim_95_A6"] == approx(plsmodel.T2_limit(0.95), rel=1e-1)
    assert data["expected_T2_lim_99_A6"] == approx(plsmodel.T2_limit(0.99), rel=1e-1)

    assert np.mean(
        np.abs(data["expected_T"].values) - np.abs(plsmodel.x_scores.values)
    ) == approx(0, abs=1e-4)
    assert np.mean(
        np.abs(data["expected_P"].values) - np.abs(plsmodel.x_loadings.values)
    ) == approx(0, abs=1e-5)
    assert np.mean(
        np.abs(data["expected_W"].values) - np.abs(plsmodel.x_weights.values)
    ) == approx(0, abs=1e-6)
    assert np.mean(
        np.abs(data["expected_C"].values) - np.abs(plsmodel.y_loadings.values)
    ) == approx(0, abs=1e-6)
    assert np.mean(
        np.abs(data["expected_U"].values) - np.abs(plsmodel.y_scores.values)
    ) == approx(0, abs=1e-5)
    assert np.mean(
        data["expected_Hotellings_T2_A3"].values.ravel()
        - plsmodel.Hotellings_T2.iloc[:, 2].values.ravel()
    ) == approx(0, abs=1e-6)
    assert np.mean(
        data["expected_Hotellings_T2_A6"].values.ravel()
        - plsmodel.Hotellings_T2.iloc[:, 5].values.ravel()
    ) == approx(0, abs=1e-6)
    assert np.mean(
        data["expected_SD_t"].ravel()
        - plsmodel.scaling_factor_for_scores.values.ravel()
    ) == approx(0, abs=1e-5)

    # Absolute sum of the deviations, accounting for the fact that each column in Y has quite
    # different range/scaling.
    assert np.sum(
        np.abs(
            np.sum(
                np.abs(
                    Y_mcuv.inverse_transform(plsmodel.y_predicted)
                    - data["expected_Yhat_A6"].values
                )
            )
            / Y_mcuv.center_
        )
    ) == approx(0, abs=1e-2)


def test_PLS_SIMCA_LDPE_missing_data(fixture_PLS_LDPE_example):
    """Unit test for LDPE case study.
    From visual inspection, observation 12 has low influence in the model.
    Set 1 value in this observation to missing and check that the results are similar to the
    full-data case: "test_PLS_SIMCA_LDPE",
    the only differences are that the tolerances are slightly relaxed.

    """
    data = fixture_PLS_LDPE_example
    data["X"].iloc[11, 0] = np.NaN
    plsmodel = PLS(n_components=data["A"], missing_data_settings=dict(md_method="scp"))

    X_mcuv = MCUVScaler().fit(data["X"])
    Y_mcuv = MCUVScaler().fit(data["Y"])
    plsmodel = plsmodel.fit(X_mcuv.transform(data["X"]), Y_mcuv.transform(data["Y"]))
    # Can only get these to very loosely match
    assert data["expected_T2_lim_95_A6"] == approx(plsmodel.T2_limit(0.95), rel=1e-1)
    assert data["expected_T2_lim_99_A6"] == approx(plsmodel.T2_limit(0.99), rel=1e-1)

    assert np.mean(
        np.abs(data["expected_T"].values) - np.abs(plsmodel.x_scores.values)
    ) == approx(0, abs=1e-2)
    assert np.mean(
        np.abs(data["expected_P"].values) - np.abs(plsmodel.x_loadings.values)
    ) == approx(0, abs=1e-3)
    assert np.mean(
        np.abs(data["expected_W"].values) - np.abs(plsmodel.x_weights.values)
    ) == approx(0, abs=1e-3)
    assert np.mean(
        np.abs(data["expected_C"].values) - np.abs(plsmodel.y_loadings.values)
    ) == approx(0, abs=1e-3)
    assert np.mean(
        np.abs(data["expected_U"].values) - np.abs(plsmodel.y_scores.values)
    ) == approx(0, abs=5e-1)
    assert np.mean(
        data["expected_Hotellings_T2_A3"].values.ravel()
        - plsmodel.Hotellings_T2.iloc[:, 2].values.ravel()
    ) == approx(0, abs=1e-6)
    assert np.mean(
        data["expected_Hotellings_T2_A6"].values.ravel()
        - plsmodel.Hotellings_T2.iloc[:, 5].values.ravel()
    ) == approx(0, abs=1e-6)
    assert np.mean(
        data["expected_SD_t"].ravel()
        - plsmodel.scaling_factor_for_scores.values.ravel()
    ) == approx(0, abs=1e-2)

    # Absolute sum of the deviations, accounting for the fact that each column in Y has quite
    # different range/scaling.
    assert np.sum(
        np.abs(
            np.sum(
                np.abs(
                    Y_mcuv.inverse_transform(plsmodel.y_predicted)
                    - data["expected_Yhat_A6"].values
                )
            )
            / Y_mcuv.center_
        )
    ) == approx(0, abs=0.5)



# (c) Kevin Dunn, 2019-2021. MIT License.
import webbrowser
import numpy as np

from matplotlib import cm
from bokeh.models import ColorBar, BasicTicker, LinearColorMapper, PrintfTickFormatter
from bokeh.plotting import figure, ColumnDataSource
from bokeh.plotting import show as show_plot
from bokeh.models import HoverTool, Range1d
from bokeh.colors import RGB

from .models import predict


def get_plot_title(main, model, prefix=""):
    """
    Constructs a sensible plot title from the ``model``.
    """
    if main is not None:
        main = prefix
        title = model.get_title()
        if title:
            main += f": {title}"

    return main


def pareto_plot(
    model,
    ylabel="Effect name",
    xlabel="Magnitude of effect",
    # show all factors and interactions
    up_to_level=None,
    main="Pareto plot",
    legendtitle="Sign of coefficients",
    negative=("Negative effects", "#0080FF"),
    positive=("Positive effects", "#FF8000"),
    show=True,
    plot_width=500,
    plot_height=None,
    # In the hover over the bars
    aliasing_up_to_level=2,
):

    # TODO: show full variable names + units on pareto plot for main factors
    #       an interactions

    """
    Plots the Pareto plot for a given model.

    Parameters
    ----------
    model: required; a model created by the package.
    ylabel: string; optional, default: "Effect name"
        The label on the y-axis of the Pareto plot.
    xlabel: string; optional, default: "Magnitude of effect"
        The label on the x-axis of the Pareto plot
    up_to_level: integer, default = None [all levels]
        Up to which level interactions should be displayed
    main: string; optional, default: "Pareto plot"
        The plot title.
    legendtitle: string; optional, default: "Sign of coefficients"
        The legend's title.
    negative: tuple; optional, default: ("Negative", "grey")
        The first entry is the legend text for negative coefficients, and the
        second entry is the colour of the negative coefficient bars.
    positive: tuple; optional, default: ("Positive", "black")
        The first entry is the legend text for positive coefficients, and the
        second entry is the colour of the positive coefficient bars.
    show: boolean; optional, default: True
        Whether or not to show the plot directly.
    aliasing_up_to_level: int; optional, default: 2
        Shows aliasing as hover entries on the bars, up to this level of
        interaction.

    Returns
    -------
    The plot handle. Can be further manipulated, e.g. for saving.

    Example
    -------

    model = linear(...)
    pareto_plot(model, main="Pareto plot for my experiment")

    p = pareto_plot(model, main="Pareto plot for my experiment", show=False)
    p.save('save_plot_to_figure.png')
    """
    # TODO:BokehDeprecationWarning: 'legend' keyword is deprecated, use explicit 'legend_label',
    #                               'legend_field', or 'legend_group' keywords instead
    # TODO: show error bars : see Bokeh annotations: Whiskers model
    # p.add_layout(
    # Whisker(source=e_source, base="base", upper="upper", lower="lower")
    # )
    # error_bars = model._OLS.conf_int()
    # http://holoviews.org/reference/elements/bokeh/ErrorBars.html
    # https://docs.bokeh.org/en/latest/docs/user_guide/annotations.html

    params = model.get_parameters()
    if up_to_level:
        assert isinstance(up_to_level, int), (
            "Specify an integer value for " "`up_to_level`."
        )
        keep = []
        for k in range(up_to_level):
            keep.extend(model.get_factor_names(level=k + 1))

        params = params.filter(keep)

    param_values = params.values
    beta_str = [f"+{i:0.4g}" if i > 0 else f"{i:0.4g}" for i in param_values]
    bar_colours = [negative[1] if p < 0 else positive[1] for p in param_values]
    bar_signs = ["Positive" if i > 0 else "Negative" for i in param_values]

    # Show the absolute parameter values, but we will colour code them
    params = params.abs()
    base_parameters = model.get_factor_names(level=1)
    full_names = []
    for param_name, _ in params.iteritems():
        if param_name in base_parameters:
            fname = model.data.pi_source.get(param_name, param_name)
            full_names.append(fname)
        else:
            full_names.append(f"Interaction between {param_name}")

    # Shuffle the collected information in the same way
    sort_order = params.argsort().values
    beta_str = [beta_str[i] for i in sort_order]
    bar_colours = [bar_colours[i] for i in sort_order]
    bar_signs = [bar_signs[i] for i in sort_order]
    full_names = [full_names[i] for i in sort_order]

    TOOLTIPS = [
        ("Short name", "@factor_names"),
        ("Full name", "@full_names"),
        ("Magnitude and sign", "@original_magnitude_with_sign"),
    ]

    alias_strings = model.get_aliases(
        aliasing_up_to_level, drop_intercept=True, websafe=True
    )

    if len(alias_strings) != 0:
        TOOLTIPS.append(
            ("Aliasing", "@alias_strings{safe}"),
        )
    else:
        alias_strings = [""] * len(params.values)
    alias_strings = [alias_strings[i] for i in sort_order]

    # And only right at the end you can sort the parameter values:
    params = params.sort_values(na_position="last")

    source = ColumnDataSource(
        data=dict(
            x=params.values,
            y=np.arange(1, len(params.index) + 1),
            factor_names=params.index.values,
            bar_colours=bar_colours,
            bar_signs=bar_signs,
            full_names=full_names,
            original_magnitude_with_sign=beta_str,
            alias_strings=alias_strings,
        )
    )

    p = figure(
        plot_width=plot_width,
        plot_height=plot_height or (500 + (len(params) - 8) * 20),
        tooltips=TOOLTIPS,
        title=get_plot_title(main, model, prefix="Pareto plot"),
    )
    p.hbar(
        y="y",
        right="x",
        height=0.5,
        left=0,
        fill_color="bar_colours",
        line_color="bar_colours",
        legend="bar_signs",
        source=source,
    )

    p.xaxis.axis_label_text_font_size = "14pt"
    p.xaxis.axis_label = xlabel
    p.xaxis.major_label_text_font_size = "14pt"
    p.xaxis.axis_label_text_font_style = "normal"
    p.xaxis.bounds = (0, params.max() * 1.05)

    p.yaxis.major_label_text_font_size = "14pt"
    p.yaxis.axis_label = ylabel
    p.yaxis.axis_label_text_font_size = "14pt"
    p.yaxis.axis_label_text_font_style = "normal"

    locations = source.data["y"].tolist()
    labels = source.data["factor_names"]
    p.yaxis.ticker = locations
    p.yaxis.major_label_overrides = dict(zip(locations, labels))

    p.legend.orientation = "vertical"
    p.legend.location = "bottom_right"

    if show:
        show_plot(p)
    else:
        return p


paretoPlot = pareto_plot


def contour_plot(
    model,
    xlabel=None,
    ylabel=None,
    main=None,
    xlim=(-3.2, 3.2),
    ylim=(-3.2, 3.2),
    colour_function="terrain",
    show=True,
    show_expt_data=True,
    figsize=(10, 10),
    dpi=100,
    other_factors=None,
):
    """
    Show a contour plot of the model.

    TODO:
    * two axes; for the real-world and coded units
    * Hover display of experimental data points
    * add a bit of jitter to the data if the numbers are exactly the same [option]


    NOTE: currently only works for variables with 2 factors. Check back for an update.

    """
    """
    valid.names <- colnames(model.frame(lsmodel))[dim(model.frame(lsmodel))[2]:2]
    if (!is.character(xlab)) {
        stop("The \"xlab\" input must be a character (string) name of a variable in the model.")
    }
    if (!is.character(ylab)) {
        stop("The \"ylab\" input must be a character (string) name of a variable in the model.")
    }
    if (!(xlab %in% valid.names)) {
        stop(paste("The variable \"", toString(xlab), "\" was not a variable name in the ",
            "linear model.\n Valid variable names are: ", toString(valid.names),
            sep = ""))
    }
    if (!(ylab %in% valid.names)) {
        stop(paste("The variable \"", toString(ylab), "\" was not a variable name in the ",
            "linear model.\n Valid variable names are: ", toString(valid.names),
            sep = ""))
    }
    """
    plt = contour_plot_bokeh(
        model,
        xlabel,
        ylabel,
        main,
        xlim,
        ylim,
        colour_function,
        show,
        show_expt_data,
        figsize,
        dpi,
        other_factors,
    )
    return plt

    # Matplotlib version
    if False:
        import matplotlib.pyplot as plt

        N = 25  # was a function input
        h_grid = np.linspace(xlim[0], xlim[1], num=N)
        v_grid = np.linspace(ylim[0], ylim[1], num=N)
        H, V = np.meshgrid(h_grid, v_grid)
        h_grid, v_grid = H.ravel(), V.ravel()

        pure_factors = model.get_factor_names(level=1)
        if xlabel is None:
            xlabel = pure_factors[0]
        else:
            xlabel = str(xlabel)

        if ylabel is None:
            ylabel = pure_factors[1]
        else:
            ylabel = str(ylabel)

        kwargs = {xlabel: h_grid, ylabel: v_grid}
        if other_factors is not None and isinstance(other_factors, dict):
            kwargs = kwargs.update(other_factors)

        # Look at which factors are included, and pop them out. The remaining
        # factors are specified at their zero level

        unspecified_factors = [i for i in pure_factors if i not in kwargs.keys()]
        for factor in unspecified_factors:
            kwargs[factor] = np.zeros_like(h_grid)

        assert sorted(kwargs.keys()) == sorted(
            pure_factors
        ), "Not all factors were specified."
        Z = predict(model, **kwargs)
        Z = Z.values.reshape(N, N)

        # Create a simple contour plot with labels using default colors.  The
        # inline argument to clabel will control whether the labels are draw
        # over the line segments of the contour, removing the lines beneath
        # the label
        _ = plt.figure(figsize=figsize, dpi=dpi, facecolor="white", edgecolor="white")
        levels = np.linspace(Z.min(), Z.max(), N)

        # Show the data from the experiment as dots on the plot
        if show_expt_data:
            plt.plot(
                model.data[xlabel],
                model.data[ylabel],
                "dimgrey",
                linestyle="",
                marker="o",
                ms=15,
                linewidth=2,
            )

        plt.title(get_plot_title(main, model, prefix="Contour plot"))
        plt.xlabel(xlabel, fontsize=12, fontweight="bold")
        plt.ylabel(ylabel, fontsize=12, fontweight="bold")

        # Set up the plot for the first time
        plt.xlim(xlim)
        plt.ylim(ylim)
        plt.grid(color="#DDDDDD")

        CS = plt.contour(H, V, Z, colors="black", levels=levels, linestyles="dotted")
        plt.clabel(CS, inline=True, fontsize=10, fmt="%1.0f")

        plt.imshow(
            Z,
            extent=[xlim[0], xlim[1], ylim[0], ylim[1]],
            origin="lower",
            cmap=colour_function,  # 'RdGy',
            alpha=0.5,
        )
        plt.colorbar()

        if show:
            plt.show()


contourPlot = contour_plot


def interaction_plot():
    """
    Interaction plot
    """
    pass


def contour_plot_bokeh(
    model,
    xlabel=None,
    ylabel=None,
    main=None,
    xlim=(-3.2, 3.2),
    ylim=(-3.2, 3.2),
    colour_function="terrain",
    show=True,
    show_expt_data=True,
    figsize=(10, 10),
    dpi=50,
    other_factors=None,
):

    # TODO: show labels of contour plot

    # https://stackoverflow.com/questions/33533047/how-to-make-a-contour-plot-in-python-using-bokeh-or-other-libs

    dpi_max = dpi ** 3.5
    N = min(dpi, np.power(dpi_max, 0.5))

    h_grid = np.linspace(xlim[0], xlim[1], num=N)
    v_grid = np.linspace(ylim[0], ylim[1], num=N)
    H, V = np.meshgrid(h_grid, v_grid)
    h_grid, v_grid = H.ravel(), V.ravel()

    pure_factors = model.get_factor_names(level=1)
    if xlabel is None:
        xlabel = pure_factors[0]
    else:
        xlabel = str(xlabel)

    if ylabel is None:
        ylabel = pure_factors[1]
    else:
        ylabel = str(ylabel)

    kwargs = {xlabel: h_grid, ylabel: v_grid}
    if other_factors is not None and isinstance(other_factors, dict):
        kwargs = kwargs.update(other_factors)

    # Look at which factors are included, and pop them out. The remaining
    # factors are specified at their zero level

    unspecified_factors = [i for i in pure_factors if i not in kwargs.keys()]
    for factor in unspecified_factors:
        kwargs[factor] = np.zeros_like(h_grid)

    assert sorted(kwargs.keys()) == sorted(pure_factors), (
        "Not all factors " "were specified."
    )
    Z = predict(model, **kwargs)
    Z = Z.values.reshape(N, N)
    z_min, z_max = Z.min(), Z.max()
    levels = np.linspace(z_min, z_max, N)

    from matplotlib.pyplot import contour, clabel
    import matplotlib.pyplot as plt
    import matplotlib

    matplotlib.use("Agg")
    # Turn interactive plotting off
    plt.ioff()
    CS = contour(H, V, Z, levels=levels, linestyles="dotted")
    clabel(CS, inline=True, fontsize=10, fmt="%1.0f")
    # contour_labels = [(float(q._x), float(q._y), float(q._text))\
    #                                                 for q in CS.labelTexts]

    # Convert the Matplotlib colour mapper to Bokeh
    # https://stackoverflow.com/questions/49931311/using-matplotlibs-colormap-for-bokehs-color-bar
    mapper = getattr(cm, colour_function)
    colours = (255 * mapper(range(256))).astype("int")
    colour_palette = [RGB(*tuple(rgb)).to_hex() for rgb in colours]
    color_mapper = LinearColorMapper(palette=colour_palette, low=z_min, high=z_max)

    # Another alternative:
    # https://stackoverflow.com/questions/35315259/using-colormap-with-bokeh-scatter
    # colors = ["#%02x%02x%02x" % (int(r), int(g), int(b)) for \
    #        r, g, b, _ in 255*mpl.cm.viridis(mpl.colors.Normalize()(radii))]

    p = figure(
        x_range=xlim,
        y_range=ylim,
        # https://github.com/bokeh/bokeh/issues/2351
        tools="pan,wheel_zoom,box_zoom,box_select,lasso_select,reset,save",
    )
    # Create the image layer
    source = {"Xax": [h_grid], "Yax": [v_grid], "predictions": [Z]}
    h_image = p.image(
        source=source,
        image="predictions",
        x=xlim[0],
        y=ylim[0],
        dw=xlim[1] - xlim[0],
        dh=ylim[1] - ylim[0],
        color_mapper=color_mapper,
        global_alpha=0.5,  # with some transparency
        name="contour_image",
    )
    h1 = HoverTool(
        tooltips=[
            (xlabel, "@{Xax}{0.4g}"),
            (ylabel, "@{Yax}{0.4f}"),
            ("Predicted", "@{predictions}{0.4g}"),
        ],
        renderers=[h_image],
        formatters={"Predicted": "printf", xlabel: "printf", ylabel: "printf"},
    )

    color_bar = ColorBar(
        color_mapper=color_mapper,
        major_label_text_font_size="8pt",
        ticker=BasicTicker(max_interval=(z_max - z_min) / N * 2),
        formatter=PrintfTickFormatter(format="%.2f"),
        label_standoff=6,
        border_line_color=None,
        location=(0, 0),
    )

    p.add_layout(color_bar, "right")

    # Contour lines using Scipy:
    # scaler_y = (ylim[1] - ylim[0]) / (N - 1)
    # scaler_x = (xlim[1] - xlim[0]) / (N - 1)
    # for level in levels:
    # contours = measure.find_contours(Z, level)
    # for contour in contours:
    # x = contour[:, 1] * scaler_y + ylim[0]
    # y = contour[:, 0] * scaler_x + xlim[0]

    for _, cccontour in enumerate(CS.allsegs):
        if cccontour:
            x = cccontour[0][:, 0]
            y = cccontour[0][:, 1]
            p.line(x, y, line_dash="dashed", color="darkgrey", line_width=1)

    # TODO: bigger experimental markers
    # TODO: hover for the data point shows the factor settings for the data point

    if show_expt_data:

        source = ColumnDataSource(
            data=dict(
                x=model.data[xlabel],
                y=model.data[ylabel],
                output=model.data[model.get_response_name()].to_list(),
            )
        )
        h_expts = p.circle(
            x="x",
            y="y",
            color="black",
            source=source,
            # linestyle='',
            # marker='o',
            size=10,
            line_width=2,
            name="experimental_points",
        )
        # custom tooltip for the experimental points
        h2 = HoverTool(
            tooltips=[
                (xlabel, "$x{0.4g}"),
                (ylabel, "$y{0.4g}"),
                ("Actual value", "@{output}{0.4g}"),  # why not working???
            ],
            renderers=[h_expts],
            formatters={"Actual value": "printf", xlabel: "printf", ylabel: "printf"},
        )
        h2.point_policy = "snap_to_data"
        h2.line_policy = "none"

    # Axis labels:
    p.xaxis.axis_label_text_font_size = "14pt"
    p.xaxis.axis_label = xlabel
    p.xaxis.major_label_text_font_size = "14pt"
    p.xaxis.axis_label_text_font_style = "bold"
    p.xaxis.bounds = (xlim[0], xlim[1])

    p.yaxis.major_label_text_font_size = "14pt"
    p.yaxis.axis_label = ylabel
    p.yaxis.axis_label_text_font_size = "14pt"
    p.yaxis.axis_label_text_font_style = "bold"
    p.yaxis.bounds = (ylim[0], ylim[1])

    # Add the hover tooltips:
    p.add_tools(h1)
    p.add_tools(h2)

    if show:
        show_plot(p)
    return p

    # Show the data from the experiment as dots on the plot
    # TODO: add some jitter

    # plt.title(get_plot_title(main, model, prefix='Contour plot'))
    # plt.xlabel(xlabel, fontsize=12, fontweight="bold")
    # plt.ylabel(ylabel, fontsize=12, fontweight="bold")

    # Set up the plot for the first time
    # plt.xlim(xlim)
    # plt.ylim(ylim)
    # plt.grid(color='#DDDDDD')

    # CS=plt.contour(H, V, Z,
    # colors='black',
    # levels=levels,
    # linestyles='dotted')
    # plt.clabel(CS, inline=True, fontsize=10, fmt='%1.0f')

    # plt.imshow(Z, extent=[xlim[0], xlim[1], ylim[0], ylim[1]],
    # origin='lower',
    # cmap=colour_function,  # 'RdGy',
    # alpha=0.5)
    # plt.colorbar()


def tradeoff_table(show_in_browser=True, show_pdf=False):
    """
    Shows the trade-off table in a web-browser. The PDF display is not
    supported yet.
    """
    # TOCONSIDER: show the image inline, esp for Jupyter notebooks.
    # https://github.com/bokeh/bokeh/issues/2426
    if show_in_browser:
        # Wrapping the image in HTML does not work in Jupyter notebooks.
        # fname = "trade-off-table.png"
        # fqp = f"file://{path.drive}/{'/'.join(path.parts[1:-1])}/media/{fname}"
        url = "https://yint.org/tradeoff"
        webbrowser.open_new_tab(url)


def plot_model(
    model,
    x_column,
    y_column=None,
    fig=None,
    x_slider=None,
    y_slider=None,
    show_expt_data=True,
    figsize=(10, 10),
    dpi=100,
    **kwargs,
):
    """
    Plots a `model` object with a given model input as `x_column` against
    the model's output: `y_column`. If `y_column` is not specified, then it
    found from the `model`.

    For model's with more than 1 inputs, the `y_column` is the variable to be
    plotted on the y-axis, and then the plot type is a contour plot.
    """
    pure_factors = model.get_factor_names(level=1)
    dpi_max = dpi ** 3.5  # should be reasonable for most modern computers
    per_axis_points = min(dpi, np.power(dpi_max, 1 / len(pure_factors)))

    # `oneD=True`: the x-variable is a model input, and the y-axis is a response
    oneD = False
    if y_column and y_column not in pure_factors:
        oneD = True
        y_column = model.get_response_name()

    param_names = [
        model.get_response_name(),
    ]
    param_names.extend(model.get_factor_names())

    # Not always a great test: y = I(1/d) does not pick up that "d" is the model
    # even though it is via the term encapsulated in I(...)
    # assert x_column in param_names, "x_column must exist in the model."
    assert y_column in param_names, "y_column must exist in the model."

    xrange = model.data[x_column].min(), model.data[x_column].max()
    xdelta = xrange[1] - xrange[0]
    xlim = kwargs.get("xlim", (xrange[0] - xdelta * 0.05, xrange[1] + xdelta * 0.05))
    h_grid = np.linspace(xlim[0], xlim[1], num=per_axis_points)
    plotdata = {x_column: h_grid}

    if not oneD:
        yrange = model.data[y_column].min(), model.data[y_column].max()
        ydelta = yrange[1] - yrange[0]
        ylim = kwargs.get(
            "ylim", (yrange[0] - ydelta * 0.05, yrange[1] + ydelta * 0.05)
        )

        v_grid = np.linspace(ylim[0], ylim[1], num=per_axis_points)
        H, V = np.meshgrid(h_grid, v_grid)
        h_grid, v_grid = H.ravel(), V.ravel()
        plotdata[x_column] = h_grid
        plotdata[y_column] = v_grid

    # TODO: handle the 2D case later

    # if other_factors is not None and isinstance(other_factors, dict):
    # plotdata = kwargs.update(other_factors)

    ## Look at which factors are included, and pop them out. The remaining
    ## factors are specified at their zero level

    # unspecified_factors = [i for i in pure_factors if i not in kwargs.keys()]
    # for factor in unspecified_factors:
    # plotdata[factor] = np.zeros_like(h_grid)

    # assert sorted(kwargs.keys()) == sorted(pure_factors), ("Not all factors "
    # "were specified.")

    Z = predict(model, **plotdata)

    if not oneD:
        assert False
    else:
        plotdata[y_column] = Z
        yrange = Z.min(), Z.max()
        ydelta = yrange[1] - yrange[0]
        ylim = kwargs.get(
            "ylim", (yrange[0] - ydelta * 0.05, yrange[1] + ydelta * 0.05)
        )

    if fig:
        p = fig
        prior_figure = True
    else:
        prior_figure = False
        p = figure(
            x_range=xlim,
            y_range=ylim,
            # https://github.com/bokeh/bokeh/issues/2351
            tools="pan,wheel_zoom,box_zoom, box_select,lasso_select,reset,save",
        )

    h_line = p.line(
        plotdata[x_column],
        plotdata[y_column],
        line_dash="solid",
        color=kwargs.get("color", "black"),
        line_width=kwargs.get("line_width", 2),
    )
    y_units = model.data.pi_units[y_column]

    tooltips = [(x_column, "$x")]
    if y_units:
        tooltips.append((f"Prediction of {y_column} [{y_units}]", "$y"))
    else:
        tooltips.append((f"Prediction of {y_column}", "$y"))

    tooltips.append(("Source", model.name or ""))

    # custom tooltip for the predicted prediction line
    h1 = HoverTool(tooltips=tooltips, renderers=[h_line])
    h1.line_policy = "nearest"

    if show_expt_data:
        source = ColumnDataSource(
            data=dict(
                x=model.data[x_column],
                y=model.data[y_column],
                output=model.data[model.get_response_name()].to_list(),
            )
        )
        h_expts = p.circle(
            x="x",
            y="y",
            color="black",
            source=source,
            size=10,
            line_width=2,
            name="Experimental_points",
        )
        h2 = HoverTool(
            tooltips=[
                (x_column, "$x"),
                (y_column, "$y"),
                ("Experimental value", "@output"),
            ],
            renderers=[h_expts],
        )

        h2.point_policy = "snap_to_data"
        h2.line_policy = "none"

    # Axis labels:
    p.xaxis.axis_label_text_font_size = "14pt"
    p.xaxis.axis_label = x_column
    p.xaxis.major_label_text_font_size = "14pt"
    p.xaxis.axis_label_text_font_style = "bold"

    p.yaxis.major_label_text_font_size = "14pt"
    p.yaxis.axis_label = y_column
    p.yaxis.axis_label_text_font_size = "14pt"
    p.yaxis.axis_label_text_font_style = "bold"
    if prior_figure:

        # p.xaxis.bounds =
        p.x_range = Range1d(
            min(xlim[0], p.x_range.start, min(model.data[x_column])),
            max(xlim[1], p.x_range.end, max(model.data[x_column])),
        )
        p.y_range = Range1d(
            min(ylim[0], p.y_range.start, min(model.data[y_column])),
            max(ylim[1], p.y_range.end, max(model.data[y_column])),
        )

    else:
        p.x_range = Range1d(xlim[0], xlim[1])
        p.y_range = Range1d(ylim[0], ylim[1])

    # Add the hover tooltips:
    p.add_tools(h1)
    p.add_tools(h2)

    show_plot(p)
    return p



# (c) Kevin Dunn, 2010-2021. MIT License. Based on own private work over the years.

import warnings
from collections import defaultdict
from typing import Optional, List, Any

import numpy as np
import pandas as pd
import statsmodels.formula.api as smf
from statsmodels.regression.linear_model import OLS
from patsy import ModelDesc


def forg(x, prec=3):
    """
    Yanked from the code for Statsmodels / iolib / summary.py and adjusted.
    """
    if prec == 3:
        # for 3 decimals
        if (abs(x) >= 1e4) or (abs(x) < 1e-4):
            return "%9.3g" % x
        else:
            return "%9.3f" % x
    elif prec == 4:
        if (abs(x) >= 1e4) or (abs(x) < 1e-4):
            return "%10.4g" % x
        else:
            return "%10.4f" % x
    else:
        raise NotImplementedError


class Model(OLS):
    """
    Just a thin wrapper around the OLS class from Statsmodels."""

    def __init__(self, OLS_instance, model_spec, aliasing=None, name=None):
        self._OLS = OLS_instance
        self._model_spec = model_spec
        self.name = name

        # Standard error
        self.df_resid = self._OLS.df_resid
        self.df_model = self._OLS.df_model
        self.nobs = self._OLS.nobs
        # Leads to errors for size inconsistency if the data frames have
        # missing data?
        # self.rsquared = self.R2 = self._OLS.rsquared
        self.residuals = self._OLS.resid

        # Will be replaced by the "lm()" function
        self.data = None
        self.aliasing = aliasing

    def __str__(self):
        spec = ModelDesc.from_formula(self._model_spec)
        return spec.describe()

    def summary(self, alpha=0.05, print_to_screen=True):
        """
        Side effect: prints to the screen.
        """
        # Taken from statsmodels.regression.linear_model.py
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            main = "OLS Regression Results"
            if self.name:
                main += ": " + str(self.name)
            else:
                if self.data.pi_title:
                    main += ": " + str(self.data.pi_title)

            smry = self._OLS.summary(title=main)
            # print(smry)
            # Call this directly and modify the result to suppress what we
            # don't really care to show:

            smry.tables[0].pop(8)

            se = "---"
            if not (np.isinf(self._OLS.scale)):
                se = f"{np.sqrt(self._OLS.scale):.3f}"

            # Residual standard error
            smry.tables[0][7][0].data = "Residual std error"
            smry.tables[0][7][1].data = se
            # smry.tables[0][7][0].data = se
            # smry.tables[0][7][1].data = se

        return smry

    def get_parameters(self, drop_intercept=True) -> pd.DataFrame:
        """Gets the paramter values; returns it in a Pandas dataframe"""

        params = self._OLS.params.copy()
        try:
            if drop_intercept:
                params.drop("Intercept", inplace=True)
        except KeyError:
            pass

        params.dropna(inplace=True)
        return params

    def get_factor_names(self, level=1):
        """
        Gets the factors in a model which correspond to a certain level:
        1 : pure factors
        2 : 2-factor interactions and quadratic terms
        3 : 3-factor interactions and cubic terms
        4 : etc
        """
        spec = ModelDesc.from_formula(self._model_spec)
        return [term.name() for term in spec.rhs_termlist if len(term.factors) == level]

    def get_response_name(self):
        spec = ModelDesc.from_formula(self._model_spec)
        return spec.lhs_termlist[0].name()

    def get_title(self) -> str:
        """Gets the model's title, if it has one. Always returns a string."""
        return self.data.get_title()

    def get_aliases(
        self,
        aliasing_up_to_level: int = 2,
        drop_intercept: Optional[bool] = True,
        websafe: Optional[bool] = False,
    ) -> list:
        """
        Returns a list, containing strings, representing the aliases
        of the fitted effects.

        aliasing_up_to_level: up to which level of interactions shown

        drop_intercept: default is True, but sometimes it is interesting to
            know which effects are aliased with the intercept

        websafe: default is False; if True, will print the first term
            in the aliasing in bold, since that is the nominally estimated
            effect.
        """
        alias_strings: List[Any] = []
        if len(self.aliasing.keys()) == 0:
            return alias_strings

        params = self.get_parameters(drop_intercept=drop_intercept)
        for p_name in params.index.values:
            if websafe:
                aliasing = (
                    '<span style="font-size: 130%; font-weight: 700">'
                    f"{p_name}</span>"
                )
            else:
                aliasing = p_name
            suffix = ""
            for alias in self.aliasing[tuple([p_name])]:

                # Subtract "-1" because the first list entry tracks the sign
                if (len(alias) - 1) <= aliasing_up_to_level:
                    aliasing += f" {alias[0]} {':'.join(alias[1:])}"
                if (len(alias) - 1) > aliasing_up_to_level:
                    if websafe:
                        suffix = r" + <i>higher interactions</i>"
                    else:
                        suffix = " + higher interactions"

            # Finished with this parameter
            alias_strings.append(aliasing + suffix)

        # All done
        return alias_strings


# Model.__repr__ = Model.__str__


def predict(model, **kwargs):
    """
    Make predictions from the model
    """
    return model._OLS.predict(exog=dict(kwargs))


def lm(
    model_spec: str,
    data: pd.DataFrame,
    name: Optional[str] = None,
    alias_threshold: Optional[float] = 0.995,
) -> Model:
    """
    Create a linear model.
    """

    def find_aliases(model, model_desc, threshold_correlation=0.995):
        """
        Finds columns which are exactly correlated, or up to at least a level
        of `threshold_correlation`.
        Returns a dictionary of aliasing and a list of columns to keep.

        The columns to keep will be in the order checked. Perhaps this can be
        improved.
        For example if AB = CD, then return AB to keep.
        For example if A = BCD, then return A, and not the BCD column to keep.
        """
        has_variation = model.exog.std(axis=0) > np.sqrt(np.finfo(float).eps)

        # np.dot(model.exog.T, model.exog)/model.exog.shape[0]
        # Drop columns which do not have any variation
        corrcoef = np.corrcoef(model.exog[:, has_variation].T)  # , ddof=0)

        # Snippet of code here is from the NumPy "corrcoef" function. Adapted.
        c = np.cov(model.exog.T, None, rowvar=True)
        dot_product = model.exog.T @ model.exog
        try:
            d = np.diag(c)
        except ValueError:
            # scalar covariance
            # nan if incorrect value (nan, inf, 0), 1 otherwise
            return c / c
        stddev = np.sqrt(d.real)

        aliasing = defaultdict(list)
        terms = model_desc.rhs_termlist
        drop_columns = []
        counter = -1
        corrcoef = c.copy()
        for idx, check in enumerate(has_variation):
            if check:
                counter += 1

                for j, stddev_value in enumerate(stddev):
                    if stddev_value == 0:
                        pass
                    else:
                        corrcoef[idx, j] = c[idx, j] / stddev[idx] / stddev[j]

                # corrcoef = c / stddev[idx, None]
                # corrcoef = corrcoef / stddev[None, idx]

                candidates = [
                    i
                    for i, val in enumerate(np.abs(corrcoef[idx, :]))
                    if (val > threshold_correlation)
                ]
                signs = [np.sign(j) for j in corrcoef[idx, :]]
            else:
                # Columns with no variation
                candidates = [
                    i
                    for i, j in enumerate(has_variation)
                    if (j <= threshold_correlation)
                ]

            # Track the correlation signs
            signs = [np.sign(j) for j in dot_product[idx, :]]

            # Now drop out the candidates with the longest word lengths
            alias_len = [(len(terms[i].factors), i) for i in candidates]
            alias_len.sort(reverse=True)
            for entry in alias_len[0:-1]:
                drop_columns.append(entry[1])

            for col in candidates:
                if col == idx:
                    # It is of course perfectly correlated with itself
                    pass
                else:

                    aliases = [t.name() for t in terms[col].factors]
                    if len(aliases) == 0:
                        aliases = ["Intercept"]

                    key = tuple([t.name() for t in terms[idx].factors])
                    if len(key) == 0:
                        key = ("Intercept",)

                    if signs[col] > 0:
                        aliases.insert(0, "+")
                    if signs[col] < 0:
                        aliases.insert(0, "-")
                    aliasing[key].append(aliases)

        # Sort the aliases in length:
        for key, val in aliasing.items():
            alias_len = [(len(i), i) if i[1] != "Intercept" else (1e5, i) for i in val]
            alias_len.sort()
            aliasing[key] = [i[1] for i in alias_len]

        return aliasing, list(set(drop_columns))

    pre_model = smf.ols(model_spec, data=data)
    model_description = ModelDesc.from_formula(model_spec)
    aliasing, drop_columns = find_aliases(
        pre_model, model_description, threshold_correlation=alias_threshold
    )
    drop_column_names = [pre_model.data.xnames[i] for i in drop_columns]

    post_model = smf.ols(model_spec, data=data, drop_cols=drop_column_names)

    name = name or data.pi_title
    out = Model(
        OLS_instance=post_model.fit(),
        model_spec=model_spec,
        aliasing=aliasing,
        name=name,
    )
    out.data = data

    return out


def summary(
    model: Model,
    show: Optional[bool] = True,
    aliasing_up_to_level: int = 3,
):
    """
    Prints a summary to the screen of the model.

    Appends, if there is any aliasing, a summary of those aliases,
    up to the (integer) level of interaction: `aliasing_up_to_level`.
    """
    out = model.summary()
    extra = []
    aliases = model.get_aliases(aliasing_up_to_level, drop_intercept=False)
    values = model.get_parameters(drop_intercept=False).values
    if len(aliases):
        extra.append("Aliasing pattern")
        for value, alias in zip(values, aliases):
            extra.append(f" {forg(value, 4)} = {alias}")

    out.add_extra_txt(extra)
    if show:
        print(out)
    return out



import pytest
from pytest import approx
import numpy as np
import pandas as pd

from process_improve.regression.methods import (
    multiple_linear_regression,
    repeated_median_slope,
    simple_robust_regression,
)


@pytest.fixture
def repeated_median():
    x = np.array([0, 1, 2, 3])
    y = np.array([5, 1, 6, 72])

    # TODO: add a test where the inputs are Pandas Series. Should handle this case also.
    # TODO: handle cases where there are nans in the vectors

    divzero_x = np.array([2, 2, 3, 4])
    divzero_y = np.array([0, 1, 2, 3])
    return x, y, divzero_x, divzero_y


def test__repeated_median(repeated_median):
    """
    (1-5)/1 ; (6-5)/2 ; (72-5)/3   => 0.5
    (1-5)/-1 ; (6-1)/1 ; (72-1)/3   => 5
    (5-6)/-2 ; (1-6)/1 ; (72-6)/2   => 5
    (5-72)/-3 ; (1-72)/-2 ; (6-72)/1   => something (23.67)
    -------
    overall median of (0.5, 5, 5 and something) is (5+5)/2 = 5
    """
    x, y, *_ = repeated_median
    assert repeated_median_slope(x, y) == 5.0


def test__repeated_median_catch_division_by_zero(repeated_median):
    """
    Ensures division by zero does not crash the algorithm
    """
    *_, divzero_x, divzero_y = repeated_median
    assert repeated_median_slope(divzero_x, divzero_y) == 1.0


@pytest.fixture
def multiple_linear_regression_data():
    """
    Verifies the linear regression calculation matches the output from R.
    > x = c(0.019847603, 0.039695205, 0.059542808, 0.07939041, 0.099238013,
            0.119085616, 0.138933218)
    > y = c(0.2, 0.195089996, 0.284090012, 0.37808001, 0.46638, 0.561559975,
            0.652559996)
    > summary(lm(y~x))
    Call:
    lm(formula = y ~ x)
    Residuals:
            1         2         3         4         5         6         7
    0.052417 -0.033668 -0.025843 -0.013029 -0.005904  0.008101  0.017925
    Coefficients:
                Estimate Std. Error t value Pr(>|t|)
    (Intercept)  0.06641    0.02710   2.451   0.0579 .
    x            4.08993    0.30530  13.396 4.15e-05 ***
    ---
    Signif. codes:  0 ‘***’ 0.001 ‘**’ 0.01 ‘*’ 0.05 ‘.’ 0.1 ‘ ’ 1
    Residual standard error: 0.03206 on 5 degrees of freedom
    Multiple R-squared:  0.9729, Adjusted R-squared:  0.9675
    F-statistic: 179.5 on 1 and 5 DF,  p-value: 4.148e-05
    > confint(lm(y~x))
                    2.5 %    97.5 %
    (Intercept) -0.003253309 0.1360676
    x            3.305124756 4.8747402
    > max(residuals(lm(y~x)))
    [1] 0.05241749
    > min(residuals(lm(y~x)))
    [1] -0.03366786

    >hatvalues(lm(y~x))
    0.4642857 0.2857143 0.1785714 0.1428571 0.1785714 0.2857143 0.4642857

    > cooks.distance(model)
    1           2           3           4           5           6           7
    2.161740342 0.308710432 0.085960163 0.016051677 0.004486278 0.017871649 0.252805729

    > predict(lm(y~x), data.frame(x=c(0.019847603,0.138933218)), interval="predict")
            fit        lwr       upr
    1 0.1475825 0.04784391 0.2473211
    2 0.6346346 0.53489603 0.7343732
    """
    X = np.array(
        [
            0.019847603,
            0.039695205,
            0.059542808,
            0.07939041,
            0.099238013,
            0.119085616,
            0.138933218,
        ]
    ).reshape(7, 1)
    y = np.array(
        [0.2, 0.195089996, 0.284090012, 0.37808001, 0.46638, 0.561559975, 0.652559996]
    )
    return X, y


def test_inconsistent_sizes(multiple_linear_regression_data):
    """Verifies that inconsistencies are picked up"""
    # TODO: X = 5 x 2, y = 7 x 1
    # TODO: X = 5 x 4, y = 5 x 1:  n=5, k=5: should work
    # TODO: X = 5 x 5, y = 5 x 1:  n=5, k=5+1 (with intercept): should fail
    X, y = multiple_linear_regression_data
    pass


def test_regression_model_with_intercept(multiple_linear_regression_data):
    """Can we reproduce the R output for single column X:
    * model output
    * confidence intervals
    """
    X, y = multiple_linear_regression_data
    out = multiple_linear_regression(X, y, fit_intercept=True, na_rm=False)
    assert out["N"] == 7
    assert isinstance(out["intercept"], float)
    assert len(out["coefficients"]) == 1
    assert out["conf_intervals"].shape == (1, 2)
    assert out["conf_interval_intercept"].shape == (2,)

    assert out["SE"] == approx(0.03206, abs=1e-5)
    assert out["R2"] == approx(0.9729, rel=1e-5)
    assert out["intercept"] == approx(0.06641, abs=1e-5)
    assert out["coefficients"][0] == approx(4.08993, rel=1e-6)
    assert out["conf_intervals"][0] == approx([3.30512, 4.87474], rel=1e-5)
    assert out["conf_interval_intercept"] == approx([-0.003253309, 0.1360676], rel=1e-6)
    assert np.max(out["residuals"]) == approx(0.05241749, rel=1e-7)
    assert np.min(out["residuals"]) == approx(-0.03366786, rel=1e-7)

    # Residuals:
    #        1         2         3         4         5         6         7
    # 0.052417 -0.033668 -0.025843 -0.013029 -0.005904  0.008101  0.017925
    assert out["residuals"][0] == approx(0.052417, abs=1e-6)
    assert out["residuals"][1] == approx(-0.033668, abs=1e-6)
    assert out["residuals"][5] == approx(0.008101, abs=1e-6)
    assert out["residuals"][6] == approx(0.017925, abs=1e-6)

    # Hat values:
    # 0.4642857 0.2857143 0.1785714 0.1428571 0.1785714 0.2857143 0.4642857
    assert out["leverage"][0] == approx(0.4642857, abs=1e-6)
    assert out["leverage"][1] == approx(0.2857143, abs=1e-6)
    assert out["leverage"][2] == approx(0.1785714, abs=1e-6)
    assert out["leverage"][3] == approx(0.1428571, abs=1e-6)

    # Influence (Cook's D):
    # 2.161740342 0.308710432 0.085960163 0.016051677 0.004486278 0.017871649 0.252805729
    assert out["influence"][0] == approx(2.161740342, abs=1e-6)
    assert out["influence"][1] == approx(0.308710432, abs=1e-6)
    assert out["influence"][2] == approx(0.085960163, abs=1e-6)
    assert out["influence"][6] == approx(0.252805729, abs=1e-6)

    # Prediction interval at the extremes
    assert out["pi_range"][0, 1] == approx(0.04784391, abs=1e-8)
    assert out["pi_range"][0, 2] == approx(0.2473211, abs=1e-8)
    assert out["pi_range"][-1, 1] == approx(0.53489603, abs=1e-8)
    assert out["pi_range"][-1, 2] == approx(0.7343732, abs=1e-7)


def test__regression_model_no_intercept(multiple_linear_regression_data):
    """Can  == approx( reproduce the R output for single rel 1E-9:
    > summary(lm(y~x+0))
    Call:
    lm(formula = y ~ x + 0)
    Residuals:
        Min        1Q    Median        3Q       Max
    -0.008637 -0.005542  0.000253  0.003448  0.105543
    Coefficients:
    Estimate Std. Error t value Pr(>|t|)
    x   4.7591     0.1849   25.74 2.27e-07 ***
    ---
    Signif. codes:  0 �***� 0.001 �**� 0.01 �*� 0.05 �.� 0.1 � � 1
    Residual standard error: 0.04343 on 6 degrees of freedom
    Multiple R-squared:  0.991, Adjusted R-squared:  0.9895
    F-statistic: 662.4 on 1 and 6 DF,  p-value: 2.268e-07
    > confint(lm(y~x+0))
        2.5 %  97.5 %
    x 4.306636 5.21157
    """
    X, y = multiple_linear_regression_data
    out = multiple_linear_regression(X, y, fit_intercept=False)

    assert len(out["coefficients"]) == 1
    assert out["conf_intervals"].shape == (1, 2)
    assert np.isnan(out["intercept"])

    assert out["SE"] == approx(0.04343, abs=1e-5)
    assert out["R2"] == approx(0.991, abs=1e-4)
    assert out["coefficients"][0] == approx(4.7591, rel=1e-6)
    assert out["conf_intervals"][0] == approx([4.306636, 5.21157], rel=1e-6)
    assert np.max(out["residuals"]) == approx(0.105543, abs=1e-6)
    assert np.min(out["residuals"]) == approx(-0.008637, abs=1e-6)


def test_input_pandas(multiple_linear_regression_data):
    """
    Pandas and Numpy inputs should be acceptable. Replicate a test used above, but with
    Pandas as the inputs.
    """
    X, y = multiple_linear_regression_data
    x = pd.DataFrame(X)
    y = pd.DataFrame(y)
    out = multiple_linear_regression(x, y)
    assert out["SE"] == approx(0.03206, abs=1e-5)


def test_input_transposed_vector(multiple_linear_regression_data):
    """
    Make the vector a column vector first.
    Also use a mix of Pandas (y) and Numpy (x).
    """
    X, y = multiple_linear_regression_data
    x = X.copy().T
    y = pd.DataFrame(y)

    # There is a difference with a transposed array
    with pytest.raises(
        AssertionError, match=r"N >= K: You need at least as many rows .*"
    ):
        _ = multiple_linear_regression(x, y)


def test_input_one_data_point():
    """
    Cannot work: fit a line with 1 datapoint
    """
    x = np.array([2])
    y = np.array([-5])
    out = multiple_linear_regression(x, y)
    assert out["N"] is None
    assert np.isnan(out["coefficients"])
    assert np.isnan(out["residuals"])


@pytest.fixture
def simple_robust_regression_data():
    """
    Check the simple robust regression model. Validated against a data set where the values
    were calculated by hand/Excel. Can we reproduce the manually calculated output?

    # Check against R as well (results from the robust output below all agree well.)
    ---------------------------
    R version 3.6.0 (2019-04-26) -- "Planting of a Tree"
    Copyright (C) 2019 The R Foundation for Statistical Computing
    Platform: x86_64-w64-mingw32/x64 (64-bit)

    > x <- c(0.0599125391, 0.0998562289, 0.1397959245, 0.1797268338, 0.2496405722)
    > y <- c(0.2788299918, 0.4663000107, 0.6585199833, 0.8372399807, 1.1684000492)
    > model <- lm(y ~  x)
    > summary(model)

    Call:
    lm(formula = y ~ x)

    Residuals:
             1          2          3          4          5
    -0.0010087 -0.0005354  0.0047065 -0.0035104  0.0003479

    Coefficients:
                Estimate Std. Error t value Pr(>|t|)
    (Intercept) -0.0006424  0.0037744   -0.17    0.876
    x            4.6815091  0.0236153  198.24 2.83e-07
    ---
    Residual standard error: 0.003459 on 3 degrees of freedom
    Multiple R-squared:  0.9999, Adjusted R-squared:  0.9999
    F-statistic: 3.93e+04 on 1 and 3 DF,  p-value: 2.83e-07

    > confint(model)
                    2.5 %     97.5 %
    (Intercept) -0.01265427 0.01136939
    x            4.60635459 4.75666351

    """
    X = np.array(
        [
            0.0599125391,
            0.0998562289,
            0.1397959245,
            0.1797268338,
            0.2496405722,
        ]
    ).reshape(5, 1)
    y = np.array(
        [
            0.2788299918,
            0.4663000107,
            0.6585199833,
            0.8372399807,
            1.1684000492,
        ]
    )
    return X, y


def test_regression_simple_robust(simple_robust_regression_data):
    X, y = simple_robust_regression_data
    out = simple_robust_regression(X, y)

    assert isinstance(out["intercept"], float)
    assert len(out["coefficients"]) == 1
    assert out["conf_interval_intercept"].shape == (2,)
    assert out["conf_intervals"].shape == (1, 2)

    assert out["SE"] == approx(0.003554009, abs=1e-10)
    assert out["R2"] == approx(0.999919429560, rel=1e-6)
    assert out["intercept"] == approx(-0.00218, abs=1e-5)
    assert out["coefficients"][0] == approx(4.69038, rel=1e-6)
    assert out["conf_intervals"][0] == approx([4.61316875, 4.76759488], rel=1e-8)
    assert out["conf_interval_intercept"] == approx(
        [-0.0145235437, 0.0101581595], rel=1e-8
    )
    assert out["residuals"][0:5] == approx(
        [0.0, 0.000118863, 0.005006413, -0.0035648, -0.000326859], abs=1e-9
    )


def test_simple_robust_regression_corner_case():
    """
    Tests some corner cases.
    """
    # No variation in x-space
    x = np.array([4, 4, 4, 4, 4])
    y = np.array([1, 2, 3, 4, 5])
    out = simple_robust_regression(x, y)
    assert np.isnan(out["standard_error_intercept"])
    assert np.isnan(out["standard_errors"][0])
    assert np.isnan(out["conf_intervals"][0][0])
    assert np.isnan(out["conf_intervals"][0][1])


def test_simple_regression_no_error():
    """
    Tests cases where there is perfect fit
    """
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([9, 8, 7, 6, 5])
    robust = simple_robust_regression(x, y)
    regular = multiple_linear_regression(x, y, fit_intercept=True)
    assert robust["influence"] == approx(regular["influence"])
    assert robust["influence"] == approx([0] * 5)
    assert robust["conf_intervals"][0] == approx(regular["conf_intervals"][0])
    assert robust["coefficients"] == approx(regular["coefficients"])



# (c) Kevin Dunn, 2010-2021. MIT License. Based on own private work over the years.

import pandas as pd


def distillateflow():
    """The flow rate of distillate from the top of a distillation column.

    These are actual data, taken 1 minute apart in time, of the flow rate leaving
    the top of a continuous distillation column (data are from a 31 day period
    in time).

    Dimensions
    ----------
    A data frame containing 44640 observations of 1 variable.

    Source
    ------
    http://openmv.net/info/distillate-flow


    Example
    -------

    """
    pass


def pollutant():
    """
    Water treatment example from BHH2, Ch 5, Question 19.

    Description
    -----------
    The data are from the first 8 rows of the pollutant water treatment example
    n the book by Box, Hunter and Hunter, 2nd edition, Chapter 5, Question 19.

    The 3 factors (C, T, and S) are in coded units where:
    C = -1 is chemical brand A; C = +1 is chemical brand B
    T = -1 is 72F for treatment temperature; T = +1 is 100F for the temperature
    S = -1 is No stirring; S = +1 is with fast stirring

    The outcome variable is:
    y = the pollutant amount in the discharge [lb/day].

    The aim is to find treatment conditions that MINIMIZE the amount of pollutant
    discharged each day, where the limit is 10 lb/day.

    Dimensions
    ----------
    A data frame containing 8 observations of 4 variables (C, S, T and y).

    Source
    ------
    Box, G. E. P. and Hunter, J. S. and Hunter, W. G.r, Statistics for
    Experimenters, Wiley, 2nd edition, Chapter 5, Question 19, page 232.

    Example
    -------

    """


def oildoe():
    """
    Industrial designed experiment to improve the volumetric heat capacity of
    a product.

    Description
    -----------

    Four materials: A, B, C and D are added in a blend to achieve a desired
    heat capacity, the response variable, y.

    The amounts were varied in a factorial manner for the 4 materials.

    The data are scaled and coded for confidentiality. All that may be
    disclosed is that variable C is either added ("Yes") or not added not
    added ("No").

    Dimensions
    ----------
    A data frame containing 19 observations of 5 variables (A, B, C, D, and
    the response, y).

    Source
    ------
    http://openmv.net/info/oil-company-doe
    Data from a confidential industrial source.

    Example
    -------


    """
    pass


def golf():
    """
    Full factorial experiments to maximize a golfer's driving distance.

    A full factorial experiment with four factors run by a golf enthusiast. The
    objective of the experiments was for the golfer to maximize her driving distance
    at a specific tee off location on her local golf course. The golfer considered
    the following factors:

    H = Tee height (cm)
    N = Holes: number of golf balls played for prior to experimental tee shot
    C = Club type
    T = Time of day (on the 24 hour clock)

    The data are in standard order, however the actual experiments were run in
    random order.

    Coded values for H, N, C and T should be used in the linear regression
    model analysis, with -1 representing the low value and +1 the high value.


    Dimensions
    ----------
    A data frame containing 16 observations of 4 variables (H, N, C, T) and a
    column y, as a response variable.

    Source
    ------
    A MOOC on Design of Experiments: ``Experimentation for Improvement'',
    https://learnche.org

    Example
    -------
    """
    pass


def boilingpot():
    """
    Full factorial experiments for stove-top boiling of water.

    Description
    -----------

    The data are from boiling water in a pot under various conditions. The
    response variable, y, is the time taken, in minutes to reach 90 degrees
    Celsius. Accurately measuring the time to actual boiling is hard, hence
    the 90 degrees Celsius point is used instead.

    Three factors are varied in a full factorial manner (the first 8
    observations). The data are in standard order, however the actual
    experiments were run in random order. The last 3 rows are runs close to,
    or interior to the factorial.

    Factors varied were:

    A = Amount of water: low level was 500 mL, and high level was 600 mL
    B = Lid off (low level) or lid on (high level)
    C = Size of pot used: low level was 2 L, and high level was 3 L.


    Coded values for A, B and C should be used in the linear regression model
    analysis, with -1 representing the low value and +1 the high value.

    Dimensions
    ----------
    A data frame containing 11 observations of 4 variables (A, B, C, with y as
    a response variable.

    Source
    ------
    MOOC on Design of Experiments: ``Experimentation for Improvement'',
    https://learnche.org

    Example
    -------
    """
    pass


def solar():
    """
    Solar panel example from Box, Hunter and Hunter, 2nd edition, Chapter 5,
    page 230.


    Description
    ------------
    The data are from a solar panel simulation case study.

    The original source that Box, Hunter and Hunter used is
    https://www.sciencedirect.com/science/article/abs/pii/0038092X67900515

    A theoretical model for a commercial system was made. A 2^4 factorial
    design was used (center point is not included in this dataset).

    The factors are dimensionless groups
    (https://en.wikipedia.org/wiki/Dimensionless_quantity), related to:

    A = total daily insolation,
    B = the tank capacity,
    C = the water flow through the absorber,
    D = solar intermittency coming in.

    All 4 factors are coded as -1 for the low level, and +1 for the high lever.

    The responses variables are
    y1: collection efficiency, and
    y2: the energy delivery efficiency.

    Dimensions
    ----------
    A data frame containing 16 observations of 6 variables (A, B, C, D, with
    y1 and y2 as responses.)

    Source
    ------
    Box, G. E. P. and Hunter, J. S. and Hunter, W. G., Statistics for
    Experimenters, 2nd edition, Wiley, Chapter 5, page 230.

    Example
    -------

    """
    pass


def data(dataset: str) -> pd.DataFrame:
    """
    Returns the ``dataset`` given by the string name.
    """



"""
Bivariate statistical tools:

* elbow detection in an (x,y) plot
* peaks: finding peaks, quantifying their height, width, center, area, left & right boundaries
* area under curve
"""

from typing import List, Union

import numpy as np

from ..regression.methods import fit_robust_lm


def find_elbow_point(x: np.ndarray, y: np.ndarray, max_iter=41) -> Union[int, float]:
    """
    Finds the elbow point when plotting numeric entries in `x` vs numeric values in list `y`.

    Returns the index into the vectors `x` and `y` [the vectors must have the same length], where
    the elbow point occurs.

    Using a robust linear fit, sorts the samples in X (independent variable)
    and takes sample 1:5 from the left, and samples (end-5):end and fits two
    linear regressions. Finds the angle between the two lines.
    Adds a point to each regression, so (1:6) and (end-6:end) and repeats.

    Finds the median angle, which is where it should stabilize.

    Will probably not work well on few data points. If so, try fitting a spline
    to the raw data and then repeat with the interpolated data.

    """
    start = 5
    # assert divmod(max_iter, 2)[1]  # must be odd number; to ensure we calculate the median later

    def calculate_line_length(
        x1: float, y1: float, x2: np.ndarray, y2: np.ndarray
    ) -> Union[float, np.ndarray]:
        """Returns the length of the line between 2 points (x1, y1) and (x2, y2), defined as:
        :math:`\\sqrt{(x2 - x1)^2 + (y2 - y1)^2}`
        """
        return np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

    # Ensure it is a Numpy array: pandas objects and lists are correctly handled.
    x = np.array(x.copy())
    y = np.array(y.copy())
    assert len(x) == len(y)

    # Stop if everything is missing:
    if np.isnan(np.nanmedian(x)) or np.isnan(np.nanmedian(y)):
        return -1

    # Eliminate missing values in x and y simultaneously.
    x, y = x[~(np.isnan(x) | np.isnan(y))], y[~(np.isnan(x) | np.isnan(y))]
    assert (
        len(x) > 10
    ), "Requires more than 10 values in the vectors (not including missing data)."
    idx_sort = x.argsort()
    x = x[idx_sort]
    y = y[idx_sort]
    N = x.size
    # Left and right anchor points: use median of the 5 points at start and end
    lft_x_avg = np.median(x[0:start])
    rgt_x_avg = np.median(x[-start:])

    int_x_list = []
    int_y_list = []
    lft_line_list = []  # type: List[float]
    rgt_line_list = []  # type: List[float]
    angle_list = []  # type: List[float]
    for i in np.floor(np.linspace(0, int(N / 2 - start) + 1, max_iter)):
        idx = int(start + i)
        lo_x, lo_y = x[0:idx], y[0:idx]
        hi_x, hi_y = x[-idx:], y[-idx:]
        lft = fit_robust_lm(lo_x, lo_y)
        rgt = fit_robust_lm(hi_x, hi_y)
        int_x, int_y = find_line_intersection(lft[1], lft[0], rgt[1], rgt[0])
        int_x_list.append(int_x)
        int_y_list.append(int_y)
        continue

        # The rest of this code is not reachable. Exploratory code: was used during development
        # only, and can be used to come up with alternative approaches of finding the elbow.
        if False:  # pragma: no cover
            # Left edge: (x=median(left 5 values), y = prediction from regression(x))
            lft_y = lft[0] + lft[1] * lft_x_avg

            # Right edge: (x=median(right 5 values), y = prediction from regression(x))
            rgt_y = rgt[0] + rgt[1] * rgt_x_avg

            lft_line = calculate_line_length(int_x, int_y, lft_x_avg, lft_y)
            rgt_line = calculate_line_length(int_x, int_y, rgt_x_avg, rgt_y)
            hypotenuse_line = calculate_line_length(lft_x_avg, lft_y, rgt_x_avg, rgt_y)
            lft_line_list.append(lft_line)
            rgt_line_list.append(rgt_line)

            angle = (
                np.arccos(
                    (lft_line ** 2 + rgt_line ** 2 - hypotenuse_line ** 2)
                    / (2 * lft_line * rgt_line)
                )
                * 180.0
                / np.pi
            )
            angle_list.append(angle)

    # Visualize the elbow point
    if False:
        import pandas as pd

        data = pd.DataFrame(data={"x": x, "y": y})
        ax = data.plot.scatter(x="x", y="y")

        intersections = pd.DataFrame(data={"x_int": int_x_list, "y_int": int_y_list})
        intersections.plot.scatter(x="x_int", y="y_int", ax=ax)
        pd.DataFrame(intersections.median()).T.plot.scatter(
            x="x_int", y="y_int", color="red", ax=ax
        )
        ax.grid(True)

    # Elbow point is taken as the average intersection point which is closest
    # to the raw data. Handle the case for even and odd number of data points
    if divmod(len(int_x_list), 2)[1] == 0:
        mid_idx = np.argmin((np.array(int_x_list) - np.nanmedian(int_x_list)) ** 2)
    else:
        mid_idx = np.where(int_x_list == np.nanmedian(int_x_list))
        if mid_idx[0].any():
            mid_idx = mid_idx[0][0]
        else:
            return np.nan

    mid_x = np.nanmedian(int_x_list)
    if np.isnan(mid_x):
        return np.nan

    # TODO: Could robustify it:
    # np.quantile(calculate_line_length(mid_x, mid_y, xraw, yraw), 0.05)
    mid_y = int_y_list[int(mid_idx)]
    return int(np.argmin(calculate_line_length(mid_x, mid_y, x, y)))


def find_line_intersection(m1: float, b1: float, m2: float, b2: float) -> tuple:
    """
    From Stackoverflow:
    stackoverflow.com/questions/20677795/how-do-i-compute-the-intersection-point-of-two-lines

    Returns a tuple: (x, y) where the two lines intersect, given slopes `m1` and
    `m2`, and intercepts `b1` and `b2`.
    """
    # These lines are essentially parallel!
    if np.abs(m1 - m2) < np.sqrt(np.finfo(float).eps):
        return np.nan, np.nan

    # y = mx + b
    # Derivation: Set both lines equal to find the intersection point in the x direction
    # m1 * x + b1 = m2 * x + b2
    # m1 * x - m2 * x = b2 - b1
    # x * (m1 - m2) = b2 - b1
    # Solving the above equation for x:
    x = (b2 - b1) / (m1 - m2)

    # Now solve it for y: use either line, because they are equal here:
    y = m1 * x + b1
    return x, y



from typing import Optional
import pandas as pd
import numpy as np
from scipy.stats import norm, iqr

from ..regression.methods import repeated_median_slope
from ..bivariate.methods import find_elbow_point

# General
# ------------------------------------------


def _prepare_data(
    df: pd.DataFrame, tags=None, batch_col=None, phase_col=None, age_col=None
):
    """
    General function, used for all feature extractions.

    1. Groups the ``df`` by batch firstly, and by phase, secondly.
    2. Creates the output dataframe to write the results to.
    """

    # Special case: a single series. Convert it to a dataframe
    if (
        isinstance(df, pd.Series)
        and (tags is None)
        and isinstance(df.index, pd.DatetimeIndex)
    ):
        if df.name is None:
            name = "tag"
        else:
            name = df.name

        tags = [name]
        df = pd.DataFrame(df, columns=tags)
        ##  NOT SURE WHAT THIS IS FOR df.insert(0, SETTINGS["batch_number_tag"], 1)

    # If no phase_col: add a fake column with all the same values
    if phase_col is None:
        phase_col = "__phase_grouper__"

    if phase_col not in df:
        # Cannot use "None" or 'nan'; else it won't group in the
        # groupby([batch_col, phase_col]) statement below.
        df.insert(0, phase_col, "NoPhases")

    # If no batch_col: add a fake column with all the same values
    if batch_col not in df or batch_col is None:
        batch_col = "__batch_grouper__"
        if batch_col not in df:
            # Cannot use "None" or 'nan'; else it won't group in the
            # groupby([batch_col, phase_col]) statement below.
            df.insert(0, batch_col, 1)

    if tags is None:
        tags = list(df.columns)
        if "__phase_grouper__" in tags:
            tags.remove("__phase_grouper__")
        if "__batch_grouper__" in tags:
            tags.remove("__batch_grouper__")

    # The user has provided a single tag name as a string, instead of a list of strings.
    if isinstance(tags, str):
        tags = [tags]

    # Check that all these columns actually exist in the df
    assert all(column_name in list(df) for column_name in tags)

    # First make a copy! else, it will repeatedly add these. Ensure it is
    # a unique list too.
    tags = tags.copy()
    tags.extend([batch_col, phase_col])
    if age_col and age_col in df:
        tags.append(age_col)
    seen_tags = []
    for tag in tags:
        if tag not in seen_tags:
            seen_tags.append(tag)

    tags = seen_tags

    if age_col and age_col in df:
        # Feature operations that require to know the time axis can
        # use a duplicate of the dataframe, and then the index from the
        # time axis to do their calculations.

        df_out = df[tags].copy()
        df_out.set_index(age_col, inplace=True)
        df_out.sort_index(inplace=True)
        tags.remove(age_col)
        grouped_data = df_out.groupby([batch_col, phase_col], as_index=True)

    else:
        df_out = df[tags].copy()
        grouped_data = df_out.groupby([batch_col, phase_col], as_index=True)

    # Remove the extra tags
    tags = [x for x in tags if x not in [batch_col, phase_col]]

    # Create the storage DF that will used to fill the outputs
    multiindex = df.index.unique()
    output = pd.DataFrame(index=multiindex)
    output.columns.name = "__features__"
    return (grouped_data, tags, output, df_out)


# Location-based features
# ------------------------------------------
def f_mean(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    mean

    The arithmetic mean for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.
    """
    base_name = "mean"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.mean()
    return output.rename(columns=dict(zip(tags, f_names)))[f_names]


def f_median(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    median

    The median for the given tags in ``tags``, for each unique batch in the ``batch_col``
    indicator column, and within each unique phase, per batch, of the ``phase_col`` column.
    """
    base_name = "median"
    prepared, tags, output, df_out = _prepare_data(data, tags, batch_col, phase_col)

    # If the feature is to be calculated within a specific range of time, then provide the
    # `slicer` input and the name of the `age_col` (column containing time evolution) which is
    # to be sliced.

    # `slicer` is a 2-element tuple: (start, stop): indicating the time points to be used during
    # which the feature is calculated. The bounds are inclusive.

    # if age_col and slicer:
    #     # The `age_col` is the index.
    #     rows = {}
    #     for idx, subgroup in prepared:
    #         rows[idx] = subgroup[
    #             (subgroup.index >= slicer[0]) & (subgroup.index <= slicer[1])
    #         ].median()

    #     # Transpose it!
    #     output = pd.DataFrame(data=rows).T
    #     f_names = [f"{tag}_{base_name}_{slicer[0]}_to_{slicer[1]}" for tag in tags]

    # The following is preferred. Not the above. Leaving it here, because it could be useful
    # to see how the slicer was constructed.
    # NOTE: an alternative for the above, is not to provide the extra function inputs of
    # `age_col` and `slicer`, but that the user pre-slices their own data.
    #
    # >>>  df = snip_head_of_all_batches(df, time_tag, snip_off_time=100)
    # >>>  df = snip_tail_of_all_batches(df, time_tag: str, snip_off_time=150)
    # >>>  f_mean(df, tags, batch_col, phase_col)

    output = prepared.median()
    f_names = [(tag + "_" + base_name) for tag in tags]
    return output.rename(columns=dict(zip(tags, f_names)))[f_names]


# Scale-based features
# ------------------------------------------
def f_std(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    std

    The standard deviation for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    See also: f_mad, f_iqr
    """
    base_name = "std"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.std()
    return output.rename(columns=dict(zip(tags, f_names)))


def f_iqr(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    iqr

    The InterQuartile Range (IQR) for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    The IQR is a robust variant of the standard deviation.
    The difference between the 75th percentile and the 25th percentile of a
    sample this is the 25 % trimmed range, an example of an L - estimator.

    See also: f_std, f_mad
    """
    base_name = "iqr"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.agg(iqr)
    return output.rename(columns=dict(zip(tags, f_names)))


def f_mad(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    mad

    The MEAN (not MEDIAN) Absolute Deviation for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    The mean absolute deviation (MAD) is a measure of the variability of a
    univariate sample of quantitative data. For values in a sequence
    X1, X2, ..., Xn, the ``mad`` is the mean of the absolute deviations from
    the data's mean.

    Since the mean can be biased by outliers, the MAD can also be biased. If
    an unbiased estimate is required, see `f_robust_mad`.

    See also: f_std, f_iqr, f_robust_mad
    """
    base_name = "mad"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.mad()
    return output.rename(columns=dict(zip(tags, f_names)))


def f_robust_mad(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    mad

    The MEDIAN (not MEAN) Absolute Deviation for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    In statistics, the median absolute deviation (MAD) is a robust measure of
    the variability of a univariate sample of quantitative data.

    For a univariate data set X1, X2, ..., Xn, the MAD is defined as the
    median of the absolute deviations from the data's median.

    from scipy.stats import norm as Gaussian
    c_MAD_constant = Gaussian.ppf(3/4.0)
    median = np.nanmedian(x)
    mad = np.nanmedian((np.fabs(x - median)) / c_MAD_constant)

    The constant correction factor is so that MAD agrees with standard
    deviation for normally distributed data.

    See also: f_mad, f_std, f_iqr,
    """
    c_MAD_const = norm.ppf(3 / 4.0)

    base_name = "mad_robust"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]

    assert False, "This next line of code fails. Fix it."
    output = (np.fabs(prepared - prepared.median())).median() / c_MAD_const

    return output.rename(columns=dict(zip(tags, f_names)))


# Cumulative features
# ------------------------------------------
def f_sum(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    sum

    The SUM within each tag for for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    If the x-axis (time) data are evenly-spaced, then this is directly
    proportional to the area under the trace (curve/trajectory).

    See also: f_cumsum
    """
    base_name = "sum"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.sum()
    return output.rename(columns=dict(zip(tags, f_names)))


def f_area(data: pd.DataFrame, time_tag, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    area

    The AREA of each tag for for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column against
    the time-based curve.

    The spacing of the x-axis is taken into account, so, this will produce
    accurate areas if the data are not evenly-spaced in time along the x-axis.

    The area is calculated using the trapezoidal rule.

    See also: f_sum, f_cumsum
    """
    base_name = "area"
    prepared, tags, output, _ = _prepare_data(
        data, tags, batch_col, phase_col, age_col=time_tag
    )
    f_names = [(tag + "_" + base_name) for tag in tags]

    # We will overwrite all entries in this dataframe, one-by-one
    output = prepared.sum()
    for batch_id, this_batch in prepared:
        # Average width of the base * height. For time series, the average width of the base is
        # the same as the sampling intervals. Therefore the diff of the index.
        half_base_factor = np.diff(this_batch.index)
        for tag in tags:
            # Now the sum makes sense: gets the area under the curve by adding
            # up the smaller trapezoids: consider the trapezoids as rotated by 90 degrees.
            # The parallel edges are lying vertically. Area of each one is the average of the
            # heights on the left and the right, multiplied by delta distance on the horizontal
            # index axis. Area = average(parallel lenghts) * height
            # where height = delta distance on the horizontal axis.
            area = (
                (this_batch[tag].iloc[0:-1].values + this_batch[tag].iloc[1:])
                / 2
                * half_base_factor
            ).sum()
            output.loc[batch_id][tag] = area

            # TODO: check this out still
            # Trapezoidal rule for integrated area, still add the "delta X" constant:
            # # https://en.wikipedia.org/wiki/Trapezoidal_rule
            # area = (
            #     this_batch[tag].values[0]
            #     + 2 * sum(this_batch[tag].values[1:-1])
            #     + this_batch[tag].values[-1]
            # ) / 2

    # output.add_suffix('_' + base_name)
    return output.rename(columns=dict(zip(tags, f_names)))


# Breakpoint detection:  rupture / breakpoint within a particular tag.
# ------------------------------------------
def f_rupture(data: pd.DataFrame, columns=None, batch_col=None, phase_col=None):
    """
    Feature:    rupture

    The breakpoint in a given tag in ``columns`` (usually it is 1 tag),
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.
    """
    # Handle phase detection based on 1 column for now.
    assert len(columns) == 1

    # TODO: see https://github.com/deepcharles/ruptures

    # base_name = "rupture"
    # prepared = _prepare_data(data, columns, batch_col, phase_col)
    # feature_columns = prepared['columns']
    # grouper = prepared['data']

    # import ruptures as rpt
    # import matplotlib.pyplot as plt
    # output = pd.DataFrame()
    # for batch_id, subset in grouper:
    # signal = subset[columns[0]].values
    # algo = rpt.Pelt(model="rbf").fit(signal)
    # result = algo.predict(pen=100)
    # print(result)
    # plt.show()
    # fig = rpt.display(signal, result, computed_chg_pts=result)
    # fig.save(batchid)


# Extreme features
# ------------------------------------------
def f_min(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    min

    The minimum value attained by each tag, for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    To get the time-point when the minimum occured: `f_agemin`.

    See also: f_agemin, f_max
    """
    base_name = "min"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.min()
    return output.rename(columns=dict(zip(tags, f_names)))


def f_max(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    max

    The maximum value attained by each tag, for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    To get the time-point when the maximum occured: `f_agemax`.

    See also: f_min
    """
    base_name = "max"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.max()
    return output.rename(columns=dict(zip(tags, f_names)))


def f_agemin(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    # TODO
    pass


def f_agemax(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    # TODO
    pass


def f_last(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    endpoint

    The final value attained by each tag, for the given tags in ``tags``,
    for each unique batch in the ``batch_col`` indicator column, and
    within each unique phase, per batch, of the ``phase_col`` column.

    If you want to know *how many* rows [i.e. the last row], then consider
    using the `f_count` feature.

    See also: f_sum, f_count
    """
    base_name = "endpoint"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.last()
    return output.rename(columns=dict(zip(tags, f_names)))


def f_count(data: pd.DataFrame, tags=None, batch_col=None, phase_col=None):
    """
    Feature:    count

    The index number of the final value for each tag, for the given tags
    in ``tags``, for each unique batch in the ``batch_col`` indicator column,
    and within each unique phase, per batch, of the ``phase_col`` column.

    Can be useful to get the 1-based index (it is a count!), and to then
    use that index for other calculation purposes.

    See also: f_sum, f_last
    """
    base_name = "count"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col)
    f_names = [(tag + "_" + base_name) for tag in tags]
    output = prepared.count()
    return output.rename(columns=dict(zip(tags, f_names)))


# Shape-based features
# ------------------------------------------
def f_slope(
    data: pd.DataFrame,
    x_axis_tag: str,
    tags=None,
    batch_col: Optional[str] = None,
    phase_col: Optional[str] = None,
    age_col: Optional[str] = None,
) -> pd.DataFrame:
    """
    Feature:    slope

    The slope of the given `tags` for each unique batch in the `batch_col` indicator column,
    of the `phase_col` column.

    The slope is calculated against whichever variable is given by `x_axis_tag`. If this is the
    `age_col` of the batch (i.e. time duration), ensure that `age_col` is also specified.

    """
    base_name = "slope"
    prepared, tags, output, _ = _prepare_data(data, tags, batch_col, phase_col, age_col)
    f_names = [(tag + "_" + base_name) for tag in tags]

    # We will overwrite all entries in this dataframe, one-by-one
    output = prepared.sum()
    for batch_id, this_batch in prepared:
        if x_axis_tag not in this_batch:
            this_batch.reset_index(inplace=True)
        for tag in tags:
            X = this_batch[x_axis_tag]
            output.loc[batch_id][tag] = repeated_median_slope(X, this_batch[tag])

    return output.rename(columns=dict(zip(tags, f_names)))


def cross(
    series: pd.Series,
    threshold: Optional[int] = 0,
    direction: Optional[str] = "cross",
    only_index: Optional[bool] = False,
    first_point_only: Optional[bool] = False,
) -> list:
    """
    Given a Series returns all the index values where the data values equal
    the 'threshold' value. Will first drop all missing values from the series.

    `direction`` can be 'rising' (for rising edge), 'falling' (for only falling
    edge), or 'cross' for both edges.

    If `only_index` is True (default False), then it will return the 0-based
    index where crossing occur *just after*. E.g. if the returned index is 135,
    then the crossing takes place at, or after, index 135, but before index 136.

    If the setting `first_point_only` is set to True, only the first point where
    the crossing occurs is reported. The rest are ignored. Default = all
    crossings are report (i.e. `first_point_only=False`).

    https://stackoverflow.com/questions/10475488/calculating-crossing-intercept-
    points-of-a-series-or-dataframe
    """
    # Find if values are above or bellow y-value crossing:
    series_no_na = series.dropna()
    above = series_no_na.values > threshold
    below = np.logical_not(above)
    left_shifted_above = above[1:]
    left_shifted_below = below[1:]
    x_crossings = []
    # Find indexes on left side of crossing point
    if direction == "rising":
        idxs = (left_shifted_above & below[0:-1]).nonzero()[0]
    elif direction == "falling":
        idxs = (left_shifted_below & above[0:-1]).nonzero()[0]
    else:
        rising = left_shifted_above & below[0:-1]
        falling = left_shifted_below & above[0:-1]
        idxs = (rising | falling).nonzero()[0]

    if len(idxs) and first_point_only:
        idxs = idxs[0]

    # Calculate x crossings with interpolation using formula for a straight line
    x1 = series_no_na.index.values[idxs]
    x2 = series_no_na.index.values[idxs + 1]
    y1 = series_no_na.values[idxs]
    y2 = series_no_na.values[idxs + 1]

    if only_index:
        return idxs

    try:
        x_crossings = (threshold - y1) * (x2 - x1) / (y2 - y1) + x1
    except TypeError:
        #  If it is a type that cannot be subtracted or multiplied:
        x_crossings = idxs

    return x_crossings


def f_crossing(
    data: pd.DataFrame,
    tag: str,
    time_tag: str,
    threshold: int = 0,
    direction: str = "cross",
    only_index: bool = False,
    batch_col: Optional[str] = None,
    phase_col: Optional[str] = None,
    suffix: Optional[str] = None,
) -> pd.DataFrame:
    """
    Feature:    cross

    The time (`time_tag`) value at which `tag` crosses a certain numeric
    `threshold``, either `direction='rising'`` (for rising edge), or
    `direction='falling'`' (for falling edge), or 'cross' for both edges.

    The time when the crossing occurs is found by linear interpolation
    between the indices. If you prefer the index itself, use `only_index=True`,
    but the default for that setting is `False`.

    Does this for each unique batch in the `batch_col` indicator column, and
    within each unique phase, per batch, of the `phase_col` column.

    `suffix`: what to add to the data tag, to name to this feature.

    Note: NaN is returned for a given batch and phase, if the crossing is not
    found.

    """
    if suffix is None:
        base_name = f"cross-{int(threshold)}"
    else:
        base_name = str(suffix)

    assert isinstance(tag, str)
    assert tag in data, f"Desired tag ['{tag}'] not found in the dataframe."

    prepared, tags, output, _ = _prepare_data(
        data,
        [tag],
        batch_col,
        phase_col=None,
        age_col=time_tag,
    )

    f_name = tag + "_" + base_name

    output = prepared.apply(
        lambda x: cross(
            x[tag], threshold, direction, only_index=only_index, first_point_only=True
        )
    )

    return pd.DataFrame(data={f_name: output})


def f_elbow(
    data: pd.DataFrame,
    x_axis_tag: str,
    tags=None,
    only_index: Optional[bool] = False,
    batch_col: Optional[str] = None,
    phase_col: Optional[str] = None,
) -> pd.DataFrame:
    """
    Feature:    elbow

    The "elbow" of the given ``tags`` for each unique batch in the ``batch_col`` indicator column,
    of the ``phase_col`` column.

    The elbow is calculated against whichever variable is given by `x_axis_tag` (usually a time-
    based tag).

    The function returns the *value* on the x-axis where the elbox occurs. Sometimes you might
    want the *index* of the value, so you can also find the corresponding y-axis value. Use
    `only_index=True` for such cases.
    """
    import warnings

    base_name = "elbow"
    prepared, tags, output, _ = _prepare_data(
        data, tags, batch_col, phase_col, age_col=x_axis_tag
    )
    f_names = [(tag + "_" + base_name) for tag in tags]

    # We will overwrite all entries in this dataframe, one-by-one
    output = prepared.sum()
    for batch_id, this_batch in prepared:

        if x_axis_tag not in this_batch:
            this_batch.reset_index(inplace=True)
        for tag in tags:
            subset = this_batch[[x_axis_tag, tag]]
            subset = subset.dropna()
            X = subset[x_axis_tag]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                elbow_index = find_elbow_point(X, subset[tag])
                if elbow_index < 0:
                    elbow_index = np.nan

                if only_index:
                    output.loc[batch_id][tag] = elbow_index
                else:
                    if np.isnan(elbow_index):
                        output.loc[batch_id][tag] = np.isnan
                    else:
                        output.loc[batch_id][tag] = X[elbow_index]

    return output.rename(columns=dict(zip(tags, f_names)))



import pathlib
import pytest

from pytest import approx
import numpy as np
import pandas as pd

from process_improve.monitoring.control_charts import ControlChart


class test_validate_against_R_qcc_xbar_one:
    """
    Uses testing data (on actual and simulated data) to verify the results agree with other
    another software package, the "qcc" library in R.

    R commands
    ----------
    data <- read.csv('https://openmv.net/file/rubber-colour.csv')
    chart <- qcc(data=data$Colour, type="xbar.one")
    target = chart$center   # 238.78
    s = chart$std.dev       # 10.43234
    """

    folder = (
        pathlib.Path(__file__).parents[1]
        / "process_improve"
        / "datasets"
        / "monitoring"
    )
    cc_values = pd.read_csv(folder / "rubber-colour.csv")
    y = cc_values["Colour"]

    # Do we get similar results to the "chart <- qcc(data=data$Colour, type="xbar.one")" from R?
    cc = ControlChart(variant="xbar.no.subgroup", style="regular")
    cc.calculate_limits(y)
    assert cc.target == approx(238.78, abs=1e-2)
    # cannot get this more precise, since the R code used in QCC follows a different approach:
    # calculates the std dev from a sequence of differences from point to point.
    assert round(cc.s) == 11

    cc = ControlChart(variant="xbar.no.subgroup", style="robust")
    cc.calculate_limits(y)
    assert cc.target == approx(239.5, abs=1e-1)
    # cannot get this more precise, since the R code used in QCC follows a different approach:
    # calculates the std dev from a sequence of differences from point to point.
    assert cc.s == approx(14.0847, abs=1e-3)


class test_control_chart:
    """
    Generic control chart tests.
    """

    y = np.array([33, 30, 29, 30, 27, 32, 17])
    # Simulated in R: loc=50, sd=4
    known_s = np.array(
        [41.5, 43.0, 51.8, 50.4, 45.6, 49.5, 45.3, 49.0, 41.4, 51.8, 58.4]
    )

    cc = ControlChart(
        variant="xbar.no.subgroup",
        style="robust",
    )
    cc.calculate_limits(y)
    assert cc.target == np.median(y[0:-1])
    assert np.all(y[cc.idx_outside_3S] == [17])

    # test_with_known_s(self):
    cc = ControlChart(variant="hw")
    cc.calculate_limits(known_s, target=50, s=4)
    assert cc.target == 50
    assert cc.s == 4


class test_holt_winters_control_chart:
    """
    Testing the Holt-Winters control chart settings.
    """

    y = np.array([10.09, 9.08, 3.14, 7.00, 11.47, 11.95, 3.96, 8.18, 1.87, 8.72])

    medium_length = np.array(
        [
            98.235,
            93.380,
            98.345,
            92.535,
            93.635,
            92.595,
            98.715,
            98.715,
            103.365,
            93.945,
            91.185,
        ]
    )

    short_length = np.array([86.115, 91.615])
    with_missing = np.array(
        [
            101.613,
            94.355,
            100.806,
            100.269,
            100.269,
            102.419,
            106.452,
            102.688,
            111.828,
            100.538,
            147.043,
            104.301,
            101.613,
            98.118,
            95.161,
            np.NaN,
            95.968,
            96.505,
            101.075,
            96.505,
            99.731,
            94.892,
            98.118,
            100.000,
            99.462,
        ]
    )

    # TODO: an example with NaN right at the start
    # TODO: an example with NaN in the first few samples (warm up)

    # def test_asserts:
    # Checks that the required asserts are raised

    cc = ControlChart()
    with pytest.raises(
        AssertionError, match="Lambda_1 must be less than or equal to 1.0."
    ):
        cc.calculate_limits(y, ld_1=1.2, ld_2=0.5)

    with pytest.raises(
        AssertionError, match="Lambda_1 must be greater than or equal to zero."
    ):
        cc.calculate_limits(y, ld_1=-1, ld_2=0.5)

    # test_hw_chart_missing_values(self):
    cc = ControlChart()
    cc.calculate_limits(with_missing, ld_1=0.4, ld_2=0.7)
    assert cc.target == approx(100.3, abs=1e-1)
    assert cc.s == approx(6, abs=1)

    # test_hw_short_length:
    # Ensures that short length sequences are also handled well.

    cc = ControlChart()
    cc.calculate_limits(short_length, ld_1=0.2, ld_2=0.5)
    assert cc.target == approx((86.115 + 91.615) / 2.0, abs=1e-3)
    assert cc.s == approx(np.std([86.115, 91.615], ddof=1))

    # test_hw_medium_length
    # Ensures that short length sequences are also handled well.
    cc = ControlChart()
    cc.calculate_limits(medium_length, ld_1=0.2, ld_2=0.5)
    assert cc.target == approx(93.945, abs=1e-3)
    assert cc.s == approx(4.493, abs=1e-3)
    assert len(cc.idx_outside_3S) == 0


class test_holt_winters_control_chart_BatchYield:
    """
    Testing the Holt-Winters control chart settings on a different dataset: batch yields
    http://openmv.net/info/batch-yield-and-purity (Kevin Dunn, personal data)
    """

    folder = (
        pathlib.Path(__file__).parents[1]
        / "process_improve"
        / "datasets"
        / "monitoring"
    )
    cc_values = pd.read_csv(folder / "batch-yield-and-purity.csv")
    subgroupsN = 3
    rounder = int(np.floor(cc_values.shape[0] / 3))
    subgroups = (
        cc_values["yield"].values[0 : (rounder * subgroupsN)].reshape(rounder, 3)
    )
    y = subgroups.mean(axis=1)

    cc = ControlChart()
    cc.calculate_limits(y)  # ld_1=0.5, ld_2=0.5
    assert cc.target == approx(75.1, abs=1e-1)
    assert cc.s == approx(4.16, abs=1e-3)
